
__author__ = "Andrea Tramacere"



import numpy as np

from astropy.table import Table

from .template_model import SpectralTemplateLogLog
from .output import section_separator
from .model_manager import FitModel
from .loglog_poly_model import LogParabolaEp,LogCubic,find_max_cubic,LogLinear
from .analytical_model import Disk
from .minimizer import fit_SED
from .plot_sedfit import  PlotSED

__all__=['filter_interval','find_E0','IC_fit_range','index','index_array','index_typecasting','peak_values',
         'SEDShape','spectral_index_range','sync_fit_range']

def filter_interval(x,x_range):
    msk1=x>=x_range[0]
    msk2=x<=x_range[1]
    return msk1*msk2



#----------------------------------------------------
class index_typecasting(object):
    """
    Class to handle different types of spectral indices
    """
    def __init__(self,data_type,val,error=False):
        self.spectral=None
        self.photon=None
        self.sed=None
        
         
        if error==False:
            if data_type=='spectral':
                self.spectral=val
                self.photon=val-1
                self.sed=val+1
            
            if data_type=='photon':
                self.spectral=val+1
                self.photon=val
                self.sed=val+2
    
            if data_type=='sed':
                self.spectral=val-1
                self.photon=val-2
                self.sed=val
        else:
            self.spectral=val
            self.photon=val
            self.sed=val
    

class index(object):
    """
    Class for the spectral indices
    """
    
    def __init__(self,name=None,data_type=None,val=None,err=None,idx_range=[]):
        index_names=['radio','radio_mm','mm_IR','IR_Opt','Opt_UV','BBB','X','UV_X','Fermi','TeV']

        data_type_allowed=['spectral','photon','sed']
        
        if name in index_names:
            self.name=name
        else:
            raise RuntimeError("index name=%s not allowed, allowed names=%"%(name,index_names))

        if data_type in data_type_allowed:
            self.data_type=data_type
        else:
            raise RuntimeError("idx_type =%s not allowed, possible values are=%s"%(data_type,data_type_allowed))

        if val is not None  :
            self.val=index_typecasting(self.data_type,val=val)
        else:
            self.val=None
        
        if err is not None  :
            self.err=index_typecasting(self.data_type,val=err,error=True)
        else:
            self.err=None
        
        if idx_range!=[]:
            self.idx_range=tuple(idx_range)
        else:
            self.idx_range=spectral_index_range(name)
            
    
    def assign_val(self,val,err):
        self.val=index_typecasting(self.data_type,val=val)
        self.err=index_typecasting(self.data_type,val=err,error=True)
       
    def show_val(self):
        if self.val is None:
            val='No'
        else:
            val=str("%e"%self.val.photon)
            
        if self.err is None:
            err='No'
        else:
            err=str("%e"%self.err.photon)

        range1=str("%.3f"%self.idx_range[0])
        
        range2=str("%.3f"%self.idx_range[1])
        
        print ("---> name = %-16s range=[%-6s,%-6s] log(Hz)  photon.val=%-13s, err=%-13s"%(self.name,range1,range2,val,err))

class index_array(object):
    """
    Class to handle an array of :class:`index` objects
    """
    def __init__(self):
        self.idx_array=[]
    
    def add_index(self,name=None,data_type=None,val=None,err=None,idx_range=[]):
        self.idx_array.append(index(name=name,data_type=data_type,val=val,err=err,idx_range=idx_range))
    
    def get_by_name(self,name):
        for pi in range(len(self.idx_array)):
            if self.idx_array[pi].name==name:
                return self.idx_array[pi]
        else:
            print ("no index with name %s found"%name)
            return None
            
    def show_pars(self):
        for par in self.idx_array:
            par.show_val()
#----------------------------------------------------

#----------------------------------------------------
def spectral_index_range(name):
        spectral_range_dic={}
        spectral_range_dic['radio']=[6,10]
        spectral_range_dic['radio_mm']=[10,11]
        spectral_range_dic['mm_IR']=[11,13]
        spectral_range_dic['IR_Opt']=[13,14]
        spectral_range_dic['Opt_UV']=[14,16]
        spectral_range_dic['UV_X']=[15,17.5]
        spectral_range_dic['BBB']=[15,16]
        spectral_range_dic['X']=[16,19]
        spectral_range_dic['Fermi']=[22.38,25.38]
        spectral_range_dic['TeV'] = [25.00, 28.38]
        return spectral_range_dic[name]
        
        
def sync_fit_range(name,indices):
        spectral_range_dic={}
        spectral_range_dic['blind']=[9,19]        
        spectral_range_dic['ISP']=[10,17]
        spectral_range_dic['LSP']=[10,17]
        spectral_range_dic['HSP']=[11,20]

        if name=='ISP' and indices.get_by_name('X').val is not None  :
            if  indices.get_by_name('X').val.sed<0:
                return [10,18]
        else:
            return spectral_range_dic[name]
        
      

        return spectral_range_dic[name]
        
def IC_fit_range(name):
    """

    :param name:
    :return:
    """
    spectral_range_dic={}
    spectral_range_dic['blind']=[16,28]
    spectral_range_dic['ISP']=[17,28]
    spectral_range_dic['LSP']=[17,28]
    spectral_range_dic['HSP']=[22,28]

    return spectral_range_dic[name]





class peak_values(object):
        """
        This Class is designed to store the  SED peak values
    
        
        """
        
        def __init__(self,name,nu_p_val=None,nu_p_err=None,nuFnu_p_val=None,nuFnu_p_err=None,curv_val=None,curv_err=None):
            """
            Constructor

            members:

                name->the name member is used to refer to Synchrortorn (S) or Inverse Compoton (IC)
                      components

                the other members are slef-esplicative
            """
            
            self.name=name 
            self.nu_p_err=nu_p_err
            self.nu_p_val=nu_p_val
            self.nuFnu_p_err=nuFnu_p_err
            self.nuFnu_p_val=nuFnu_p_val
            self.curvature=curv_val
            self.curvature_err=curv_err

            
        def update(self,fit_model,fit_law_name):
            #!!CHANGE TO GET PAR BY TYPE
            """ the show() method returns the class members values """
            #print('-->',fit_model,fit_law_name)
            #print(fit_model.show_model())
            Ep=fit_model.parameters.get_par_by_name(fit_law_name,'Ep')
            Sp=fit_model.parameters.get_par_by_name(fit_law_name,'Sp')
            curv=fit_model.parameters.get_par_by_name(fit_law_name,'b')
            #print('-->',Ep,type(Ep),dir(Ep))
            self.nu_p_val=Ep.best_fit_val
            self.nu_p_err=Ep.best_fit_err
            self.nuFnu_p_val=Sp.best_fit_val
            self.nuFnu_p_err=Sp.best_fit_err
            self.curvature=curv.best_fit_val
            self.curvature_err=curv.best_fit_err
            
            
        
        def show(self):
            if self.nu_p_val is None:
                nu_p_val='No'
            else:
                nu_p_val=str("%+e"%self.nu_p_val)
            
            
            if self.nu_p_err is None:
                nu_p_err='No'
            else:
                nu_p_err=str("%+e"%self.nu_p_err)
            
            if self.nuFnu_p_val is None:
                nuFnu_p_val='No'
            else:
                nuFnu_p_val=str("%+e"%self.nuFnu_p_val)
            
            
            if self.nuFnu_p_err is None:
                nuFnu_p_err='No'
            else:
                nuFnu_p_err=str("%+e"%self.nuFnu_p_err)
            
            
            if self.curvature is None:
                curvature='No'
            else:
                curvature=str("%+e"%self.curvature)
            
            
            
            if self.curvature_err is None:
                curvature_err='No'
            else:
                curvature_err=str("%+e"%self.curvature_err)
            
            """ the update() method returns update the class members"""
            
            print ("---> %-10s nu_p=%-13s (err=%-13s)  nuFnu_p=%-13s (err=%-13s) curv.=%-13s (err=%-13s)"%(self.name,nu_p_val,nu_p_err,nuFnu_p_val,nuFnu_p_err,curvature,curvature_err))
        #def update(nu_p_val=None,nu_p_err=None,nuFnu_p_val=None,nuFnu_p_err=None,curv=None,curv_err=None):
            
    
#----------------------------------------------------


#
class SEDShape(object):
    """
    
    This handle the SED shaping process

    """    
    
    def __init__(self,sed_data):
        
        self.indices=index_array()
        self.indices.add_index(name='radio',data_type='sed')    
        self.indices.add_index(name='radio_mm',data_type='sed')   
        self.indices.add_index(name='mm_IR',data_type='sed')
        self.indices.add_index(name='IR_Opt',data_type='sed')
        self.indices.add_index(name='Opt_UV',data_type='sed')
        self.indices.add_index(name='BBB',data_type='sed')
        self.indices.add_index(name='UV_X',data_type='sed')
        self.indices.add_index(name='X',data_type='sed')
        self.indices.add_index(name='Fermi',data_type='sed')
        self.indices.add_index(name='TeV', data_type='sed')
        
        self.sed_data=sed_data
        
        self.cosmo=sed_data.cosmo

       
        self.BBB=None
        self.disk=None
        self.L_host=None
        self.L_host_err=None
        
        self.L_Disk=None
        self.L_Disk_err=None
        
        
        self.T_Disk=None
        
        self.nu_p_Disk=None
        self.nu_p_Disk_err=None
       
        self.S_nu_max=None
        self.IC_nu_max=None
        
        self.obj_class=None
        
        
        

        self.S_peak=peak_values(name='sync')
        self.S_LE_slope=None
        
        self.IC_peak=peak_values(name='IC')
        self.IC_LE_slope=None

        self.host_gal=None
        self.sync_fit_model=None
        self.IC_fit_model=None

    def find_class(self,E_S):
        """

        method to evaluate obj class 'L/I/HPS' according
        to Ep
        
        args: E_S
        
        the obj_class member is updated
        
        obj_class==None means undefined class

        """

        if E_S>6 and E_S<14:
     
            self.obj_class='LSP'
            print ("--> class: ", self.obj_class)
        elif E_S>=14 and  E_S<=15:
             
            self.obj_class='ISP'
            print ("---> class: ", self.obj_class)
        elif E_S>15:
             
            self.obj_class='HSP'
            print ("---> class: ", self.obj_class)
        else:
            self.obj_class=None
            print ("---> undefined class for src",self.obj_class)
    
    
    def show_values(self):
        
        print (section_separator)
        
        print ('*** SEDShape values ***')

        print ('---> spectral inidces values')
        
        for index in self.indices.idx_array:
        
            index.show_val()
        
        print()
        print()
        print ("---> S/IC peak values")
        
        self.S_peak.show()
        
        print ()
        
        self.IC_peak.show()

        print ()
        print ()

        if self.L_Disk is not None  :
            
            print ("---> Disk peak values")
            print ("---> %-10s nu_p_Diks=%+13e (err=%+13e)  L_Disk=%+13e (err=%+13e) T_Disk=%+13e "%('Disk',self.nu_p_Disk,
                                                                                                           self.nu_p_Disk_err,
                                                                                                           self.L_Disk,
                                                                                                           self.L_Disk_err,
                                                                                                           self.T_Disk))
        
        print (section_separator)
    
    
    
    def save_values(self,name):

        _v=[]

        _dt=[('src_name','S32')]
        _v = [name]
        for index in self.indices.idx_array:
            _dt.append((index.name, 'f8'))
            _dt.append((index.name + '_err', 'f8'))
            if index.val is not None:

                _v.extend([index.val.photon, index.err.photon])
            else:
                _v.extend([None,None])



        _dt.append(('nu_p_S', 'f8'))
        _dt.append(('nu_p_S_err', 'f8'))
        _v.extend([self.S_peak.nu_p_val,self.S_peak.nu_p_err])

        _dt.append(('nuFnu_p_S', 'f8'))
        _dt.append(('nuFnu_p_S_err', 'f8'))
        _v.extend([self.S_peak.nuFnu_p_val, self.S_peak.nuFnu_p_err])

        _dt.append(('nu_p_IC', 'f8'))
        _dt.append(('nu_p_IC_err', 'f8'))
        _v.extend([self.IC_peak.nu_p_val, self.IC_peak.nu_p_err])

        _dt.append(('nuFnu_p_IC', 'f8'))
        _dt.append(('nuFnu_p_IC_err', 'f8'))
        _v.extend([self.IC_peak.nuFnu_p_val, self.IC_peak.nuFnu_p_err])


        _out=np.zeros(1,dtype=_dt)
        _out[0]=tuple(_v)

        t=Table(_out)
        t.write(name,overwrite=True)

    def eval_indices(self,minimizer='lsb',silent=True,show_fit_report=False):
        """
        
        This methods evaluates the indices for the SED
        indices are istances of the index_array () class
        
        """
        
        
        print (section_separator)
        
        print ("*** evaluating spectral indices for data ***")
        
        self.index_models=[]
        
        for index in self.indices.idx_array:
            do_fit=self.check_adapt_range_size(self.sed_data.data['nu_data_log'],index,3,silent=silent)
            if do_fit==True:
                loglog_poly=LogLinear()
                loglog_pl=FitModel(cosmo=self.cosmo, name='%s'%index.name,loglog_poly=loglog_poly)
                #print(10.**index.idx_range[0],10.**index.idx_range[1])
                mm,best_fit=fit_SED(loglog_pl,
                                 self.sed_data,
                                 10.**index.idx_range[0],
                                 10.**index.idx_range[1],
                                 loglog=True,
                                 silent=silent,
                                 fitname='spectral-indices-best-fit',
                                 minimizer=minimizer,
                                 use_UL=True)
                #val,err=do_linear_fit(self.SEDdata.nu_data_log,self.SEDdata.nuFnu_data_log,dy=self.SEDdata.dnuFnu_data_log,x_range=index.idx_range)




                par=loglog_pl.parameters.get_par_by_name(loglog_poly.name,'alpha')

                index.assign_val(val = par.best_fit_val, err =par.best_fit_err)

                if silent is False:
                    index.show_val()

                self.index_models.append(loglog_pl)

                if show_fit_report==True and silent is False:
                    best_fit.show_report()
                    print()

            if silent is False:
                print()

        print (section_separator)
            

    def plot_indices(self,plot_obj=None):
        if plot_obj is None:
            plot_obj=PlotSED(sed_data=self.sed_data)

        for model in self.index_models:
            plot_obj.add_model_plot(model,label=model.name,line_style='--')
        
        return plot_obj

    def plot_shape_fit(self, plot_obj=None):
        if plot_obj is None:
            plot_obj = PlotSED(sed_data=self.sed_data)

            if self.sync_fit_model is not None:
                plot_obj.add_model_plot(self.sync_fit_model, label='sync, poly-fit')

            if self.host_gal is not None:
                plot_obj.add_model_plot(self.host_gal, label='host-gal')

            if self.BBB is not None:
                plot_obj.add_model_plot(self.BBB, label='BBB')

            if self.disk is not None:
                plot_obj.add_model_plot(self.disk, label='disk')

            #plot_obj.add_model_plot(self.sync_fit_model, label='sync+host, poly-fit')
            if self.IC_fit_model is not None:
                plot_obj.add_model_plot(self.IC_fit_model, label='IC, poly-fit')

        return plot_obj

    def sync_fit(self,check_host_gal_template=False,
                 check_BBB_template=False,
                 check_disk=False,
                 fit_range=None,
                 nu_min=None,
                 nu_max=None,
                 Ep_start=None,
                 use_log_par=False,
                 minimizer='lsb',
                 silent=True,
                 show_fit_report=False):
        
        """
        This method analyses the synchrotron shape by means
        of log-log polynomial fits

        The following paremeters are estimated:

         1) the SED peak frequency Ep
         2) the SED peak flux Sp
         3) the curvature at the peak
         4) checks for the host galaxy
         
         -) first a log-log cubic fit is performed, with the 'blind' interval

         -) the SED class 'I/L/HPS' is set according to Ep, by find_class

         -) the fit range is changed from 'blind', according to the
            value returned by find_class, using the function sync_fit_range

         -) the SED nu, and nuFnu are generated for the 'blind' fit

         -) if the option is selected in the call of the method
            the estimate of the host galaxy is performed            

         -) a second run improve the obj class is performed

        
        values are stored in the class :class: 'peak_values'

        """
        
        print  (section_separator)
                  
        print ("*** Log-Polynomial fitting of the synchrotron component ***")
        
    
       
        if use_log_par==True:
            fit_law = LogParabolaEp()
        else:
            fit_law=LogCubic()

        self.sync_fit_model=FitModel(cosmo=self.cosmo, loglog_poly=fit_law,name='sync_model')
        self.sync_fit_model.set(fit_law.name,'b',fit_range_max=0)
        self.sync_fit_model.set(fit_law.name,'b',val_max=0)
        self.sync_fit_model.set(fit_law.name,'b',fit_range_min=-10)
        self.sync_fit_model.set(fit_law.name,'b',val_min=-10)

        if Ep_start is not None  :
            self.sync_fit_model.set(fit_law.name,'Ep',val=Ep_start)

        mm, sync_best_fit = self.do_sync_fit(self.sync_fit_model,
                                             fit_law.name,
                                             fit_range=fit_range,
                                             check_disk=check_disk,
                                             check_BBB=check_BBB_template,
                                             check_host=check_host_gal_template,
                                             # use_log_par=False,
                                             Ep_start=Ep_start,
                                             minimizer=minimizer,
                                             silent=silent,
                                             show_fit_report=show_fit_report)
        #print "bbb", self.sync_fit_model.SED.nu

        #Ep=self.S_peak.nu_p_val

        #logparpl=LogParabolaPL()
        #fit_model=FitModel(loglog_poly=logparpl,name='sync_logpar_pl_model')
        #self.do_sync_fit(fit_model,fit_range=fit_range,
        #                 check_disk=check_disk,
        #                 check_BBB=check_BBB_template,
        #                 check_host=check_host_gal_template,
        #                 use_log_par=True,
        #                 Ep=Ep)



        #print "bbb", self.sync_fit_model.SED.nu

        self.sync_best_fit=sync_best_fit

        return mm,sync_best_fit
       
    
    def save_sync_fit_report(self,name=None):
        self.sync_best_fit.save_report(name=name)

    
    def do_sync_fit(self,
                    fit_model,
                    fit_law_name,
                    fit_range=None,
                    check_disk=False,
                    check_BBB=False,
                    check_host=False,
                    #use_log_par=False,
                    Ep_start=None,
                    no_check=False,
                    minimizer='lsb',
                    silent=True,
                    show_fit_report=True):
        
       
        
      
        #fit_model1=copy.deepcopy(fit_model)

        if fit_range is None:
            s_fit_range=sync_fit_range('blind',self.indices)
        else:
            s_fit_range=fit_range
            
        print ("---> first blind fit run,  fit range:",s_fit_range)

        if silent == False:
            fit_model.show_pars()

        if Ep_start is None:
            Ep=find_max_cubic(self.sed_data.data['nu_data_log'],self.sed_data.data['nuFnu_data_log'] ,x_range=s_fit_range)

        else:
            Ep=Ep_start


        #if use_log_par == False:
        fit_model.set(fit_law_name,'Ep', val=Ep)

        #else:
        #    fit_model1.set('E0', val=Ep - 1)

        mm,best_fit=fit_SED(fit_model,self.sed_data,10.**s_fit_range[0],10.**s_fit_range[1],loglog=True,silent=silent,fitname='sync-shape-fit',minimizer=minimizer,use_UL=True)

        self.S_peak.update(fit_model,fit_law_name)
        self.find_class(self.S_peak.nu_p_val)

        refit=False
        if check_disk==True : 
            self.add_disk(fit_model)
            refit=True
        if check_BBB==True:
            self.add_BBB_template(fit_model)
            refit=True
        if check_host==True:
            self.add_host_template(fit_model)
            refit=True

        if silent == False:
            if show_fit_report == True:
                best_fit.show_report()
        print()
        #Ep=None



        if refit==True:
            if fit_range is None:
                s_fit_range = sync_fit_range(self.obj_class, self.indices)

            mm,best_fit=fit_SED(fit_model,self.sed_data,10.**s_fit_range[0],10.**s_fit_range[1],loglog=True,silent=True,fitname='sync-shape-fit',minimizer=minimizer,use_UL=True)

            if silent == False:
                best_fit.show_report()

            #if use_log_par==False:
            self.S_peak.update(fit_model,fit_law_name)

            self.find_class(self.S_peak.nu_p_val)



        print()
        print()
        fit_model.show_best_fit_pars()
        self.S_peak.show()

        self.S_nu_max= self.get_nu_max(self.sed_data.data['nu_data_log'] ,s_fit_range)
        
        
        self.set_S_LE_slope(fit_model,fit_law_name,use_log_par=False)


        if check_disk==True :
            self.disk.set_disk_pars(fit_model,'Disk')
            self.L_Disk=self.disk.L_Disk
            self.L_Disk_err=self.disk.L_Disk_err
            self.T_Disk=self.disk.T_Disk
            self.T_Disk_err=self.disk.T_Disk_err
            self.nu_p_Disk=self.disk.nu_p_Disk
            self.nu_p_Disk_err=self.disk.nu_p_Disk_err


        if check_host==True :
            self.host_gal.set_host_pars(fit_model,'host_galaxy')

        if check_BBB==True :
            self.BBB.set_BBB_pars(fit_model,'BBB')
            self.L_Disk=self.BBB.nuLnu_p_BBB
            self.L_Disk_err=self.BBB.nuLnu_p_BBB_err
            self.T_Disk=self.BBB.T_Disk
            self.T_Disk_err=self.BBB.T_Disk_err
            self.nu_p_Disk=self.BBB.nu_p
            self.nu_p_Disk_err=self.BBB.nu_p_err

        print(section_separator)

        return mm,best_fit
        
    
    def add_disk(self,fit_model):
        
        disk=Disk(cosmo=fit_model.cosmo, z=self.sed_data.z,name='Disk')
        
        fit_model.add_component(disk)

        fit_model.set('Disk','nuFnu_p',val=(10**self.S_peak.nuFnu_p_val),fit_range=[10**(self.S_peak.nuFnu_p_val-2),10**(self.S_peak.nuFnu_p_val+2 )])
        
        fit_model.set('Disk','T_Disk',val=1E4,fit_range=[5000,1E5])
 
        self.disk=disk
 
 
 
 
    
    def   add_BBB_template(self,fit_model):
        
        BBB_template=SpectralTemplateLogLog.template_factory('BBB', cosmo=fit_model.cosmo, z=self.sed_data.z, name='BBB')
        
        fit_model.add_component(BBB_template)
   
        fit_model.set('BBB','nuFnu_p_BBB',val=(self.S_peak.nuFnu_p_val),fit_range=[self.S_peak.nuFnu_p_val-2,self.S_peak.nuFnu_p_val+2 ])
        
        fit_model.set('BBB','nu_scale',val=0,fit_range=[-0.5,0.5])
   
        self.BBB=BBB_template
    
    
    
        
    
    
        

    def add_host_template(self,fit_model):
        
        host_gal=SpectralTemplateLogLog.template_factory('host_galaxy', cosmo=fit_model.cosmo, z=self.sed_data.z, name='host_galaxy')
                 
        fit_model.add_component(host_gal)

        #print('nuFnu_p_host',self.S_peak.nuFnu_p_val)
        fit_model.set('host_galaxy','nuFnu_p_host',val=(self.S_peak.nuFnu_p_val),fit_range=[self.S_peak.nuFnu_p_val-2,self.S_peak.nuFnu_p_val+2 ])
        
        fit_model.set('host_galaxy','nu_scale',val=0,fit_range=[-0.5,0.5])
   
        self.host_gal=host_gal
    
    
    
    
    def set_S_LE_slope(self,fit_func,fit_law_name,use_log_par):
        if use_log_par==False:
            self.S_LE_slope=2.0
        else:
            self.S_LE_slope=fit_func.parameters.get_par_by_name(fit_law_name,'alpha').best_fit_val

    
    
    
    
      
   
    
    def IC_fit(self,fit_range=None,use_log_par=False,Ep_start=None,minimizer='minuit',silent=False):
    
        print  (section_separator)
              
        print ("*** Log-Polynomial fitting of the IC component ***")
        
        if fit_range is None:
            fit_range=IC_fit_range(self.obj_class)
        
        print ("---> fit range:",fit_range)
        

        


        if Ep_start is None:
            Ep=find_max_cubic(self.sed_data.data['nu_data_log'],self.sed_data.data['nuFnu_data_log'] ,x_range=fit_range)
        else:
            Ep=Ep_start


        if use_log_par==True:
            print("---> LogParabola fit")
            logpar_model = LogParabolaEp()
            fit_law_name = logpar_model.name
            self.IC_fit_model = FitModel(cosmo=self.cosmo, loglog_poly=logpar_model, name='IC_log-par_model')



            if Ep is not None:
                self.IC_fit_model.set(fit_law_name,'Ep', val=Ep)

            if silent == False:
                self.IC_fit_model.show_pars()

            mm,best_fit = fit_SED(self.IC_fit_model, self.sed_data, 10. ** fit_range[0], 10. ** fit_range[1], loglog=True,
                               silent=True, fitname='IC-shape-fit', minimizer=minimizer,use_UL=True)

            if silent == False:
                best_fit.show_report()

            best_fit_model = self.IC_fit_model
        else:
            print("---> LogCubic fit")
            cubic_model = LogCubic()
            fit_law_name=cubic_model.name
            self.IC_fit_model = FitModel(cosmo=self.cosmo, loglog_poly=cubic_model, name='IC_log-cubic_model')

            self.IC_fit_model.set(fit_law_name,'b', fit_range_max=0)
            self.IC_fit_model.set(fit_law_name,'b', val_max=0)
            self.IC_fit_model.set(fit_law_name,'b', fit_range_min=-10)
            self.IC_fit_model.set(fit_law_name,'b', val_min=-10)
            if Ep is not None:
                self.IC_fit_model.set(fit_law_name,'Ep', val=Ep)

            try:
                mm,best_fit=fit_SED(self.IC_fit_model,self.sed_data,10.**fit_range[0],10.**fit_range[1],loglog=True,silent=True,fitname='IC-shape-fit',minimizer=minimizer,use_UL=True)

                if silent == False:
                    best_fit.show_report()

                best_fit_model=self.IC_fit_model


            except Exception as e:

                print ("---> LogCubic fit failed",e)
                print ("---> try LogParabola ")
                logpar_model=LogParabolaEp()
                fit_law_name=logpar_model.name
                self.IC_fit_model=FitModel(cosmo=self.cosmo, loglog_poly=logpar_model,name='IC_log-par_model')
                if Ep is not None:
                    self.IC_fit_model.set(fit_law_name,'Ep', val=Ep)

                if silent == False:
                    self.IC_fit_model.show_pars()

                mm,best_fit=fit_SED(self.IC_fit_model,self.sed_data,10.**fit_range[0],10.**fit_range[1],loglog=True,silent=True,fitname='IC-shape-fit',minimizer=minimizer,use_UL=True)

                if silent == False:
                    best_fit.show_report()

                best_fit_model=self.IC_fit_model
            
            
            
        self.IC=best_fit_model
        
        self.IC_peak.update(best_fit_model,fit_law_name)

        print()
        print()
        self.IC_fit_model.show_best_fit_pars()

        self.IC_peak.show()
        
      
        self.IC_nu_max= self.get_nu_max(self.sed_data.data['nu_data_log'],fit_range)
        
        print  (section_separator)

        return mm,best_fit
    
    
    def get_nu_max(self,nu,fit_range):

        msk = filter_interval(nu,fit_range)
        return nu[msk].max()
        
        
    def check_adapt_range_size(self,x,index,min_size,silent=False):
        do_fit=True
        x_range=[index.idx_range[0],index.idx_range[1]]
        msk = filter_interval(x,x_range)
        x1=x[msk]
        delta=0.1
        delta_tot=0
        if silent is False:
            print("---> initial range for index %s  set to [%f,%f]" % (index.name, index.idx_range[0], index.idx_range[1]))
        while len(x1)<min_size and delta_tot<1.0:
            delta_tot+=delta
            x_range[0]-=delta
            x_range[1]+=delta
            msk = filter_interval(x,x_range)
            x1=x[msk]

        if len(x1)>=min_size:
            do_fit=True
            index.idx_range=x_range
            if silent is False:
                print ("---> range for index %s updated  to [%f,%f]"%(index.name,index.idx_range[0],index.idx_range[1] ))
        else:
            do_fit=False
            if silent is False:
                print("---> not enough data in range for index%s " % (index.name))

        return  do_fit
    
def find_E0(b,a,Ep):
    """returns the value of E0  for
    a log_par+pl distribution
    
    Args:
        b:  curvature
        a:  spectral index in the PL branch
        Ep: peak value 

    Returns:
        
        
    """
    if (b!=0.0):
        c=(3-a)/(-2*b)
        print ("Ep=",Ep)
        return Ep/(pow(10,c))
    else:
        return Ep

#----------------------------------------------------   


#from __future__ import absolute_import, division, print_function

#from builtins import (bytes, str, open, super, range,
#                      zip, round, input, int, pow, object, map, zip)

__author__ = "Andrea Tramacere"

import os
import sys
import glob

from .cosmo_tools import  Cosmo
from string import *
import math as m
from astropy.io import fits as pyfits
import numpy as n

__all__=['check_is_number','do_convert']

def check_is_number(t):
    try: 
        float(t[0])
        return True
    except ValueError:
        return False



#------ Main --------------
def main(argv=None,cmd_par=None):

    if cmd_par  is None:
        argc=len(sys.argv)
        arg_list=sys.argv
        in_seds=arg_list[1:]
    else:
        argc=len(cmd_par)
        arg_list=(cmd_par)
        if '*' in arg_list[1]:
            in_seds=glob.glob(arg_list[1])

        print (in_seds)
 
    print (arg_list)
    if argc==1:
        print("./concvert_qdp.py infiles.qdq")
        print("exmaple: ./concvert_qdp.py SED_Ciprini_v1/sed-*.qdp")
        print("input data are in obs rest frame, out data in src rest frame (z-corrected)")
        sys.exit(0)

   
    print (arg_list)

    comos_eval=Cosmo(units='cm')
    for sed in in_seds:
        do_convert(sed,comos_eval)
        
def do_convert(infile_name,cosmo_eval):
    infile=open(infile_name,'r')
    
    outfile_name=infile_name.split('.')
    outfile_name[-1]='dat'
    outfile_name=".".join(outfile_name)
    outfile=open(outfile_name,'w')
    obj_name=infile_name.split('/')[-1]
    obj_name=obj_name.split('.')[0]
    obj_name=obj_name.split('-')[1]
    
    inlines=infile.readlines()
    
    infile.close()
    
    print('# set_number: -1 hyst, 0 sym, 1 green, 2 blue ',file=outfile)
    print('# file_name ',obj_name,file=outfile)
    set_code={}
    UL_sets=[]

    for line in inlines:
        line=line.strip()
        if len(line)>0:
            tkn=line.split()
            #print tkn
            if tkn[0]=="!" and tkn[1]=='Redshift':
                z=float(tkn[3])
                DL=cosmo_eval.DL(z)
                print('# z ',z,file=outfile)
                print('# DL(cm) ',DL,file=outfile)
                print('# restframe  src',file=outfile)
                print('# dataScale  log-log',file=outfile)
            
            for t in tkn:
                #print t
                if ('J' in t and '+' in t) or ('J' in t  and '-' in t):
                    name=t
                    #print 'J' and '+' in t
                    #print 'J' and '-' in t 
                    #print t
                    print('# obj_name    ',name,file=outfile)
    
            if tkn[0]=='col' and tkn[2]=='on':
                #print tkn,tkn[3:]
                for  t in tkn[3:]:
                    #print t,tkn[1],t=='2'
                    #sym
                    if tkn[1]== '2':
                        set_code[t]='0'
                        
                    #
                    elif tkn[1]=='15':
                        set_code[t]='-1'
                    #green
                    elif tkn[1]=='3':
                        set_code[t]='1'
                    #blue
                    elif tkn[1]=='4':
                        set_code[t]='2'

            #UL
            if tkn[0]=='ma' and tkn[1]=='31':
                UL_sets=tkn[3:]
                print ("UL sets",UL_sets)


    print>>outfile,'#nu nuLnu dnu dnuLnu set_number '     
    set=1
    #print set_code
    if z!=0.0:
        for line in inlines:
            print (tkn)
            tkn=line.split()           
            if  len(tkn)==2 and tkn[0]=='NO' and tkn[1]=='NO':
                #print tkn
                set+=1
                #print>>outfile,"###############################"
            if check_is_number(tkn[0]):
                if float(tkn[0])>5:
                    #print tkn
                    nu_obs=pow(10,float(tkn[0]))
                    dnu_obs=nu_obs*float(tkn[1])
                    nuLnu_obs=pow(10,float(tkn[2]))
                
                    
                    nu_rest=nu_obs*(1+z)
                    log_dnuLnu_rest=float(tkn[3])
                    #!!! ASSUMES UL in FERMI if dy==0
                    #!!! for other data sets UL from set with ma=31
                    if nu_obs>1.0e22 and log_dnuLnu_rest==0.0:
                        #print  dnu_rest
                        log_dnuLnu_rest=-1
                    
                    for UL_set in UL_sets:
                        if int(UL_set)==set:
                            #print 'UL',UL_set,set
                            log_dnuLnu_rest=-1

                    nuLnu_rest=nuLnu_obs*4*m.pi*DL*DL
                    #errore relativo 
                    #dnuLnu_rest=nuLnu_rest*(float(tkn[3]))
                        
                    print(m.log10(nu_rest),m.log10(nuLnu_rest),0,log_dnuLnu_rest,set_code['%s'%set],file=outfile)
                    #print outfile,nu_rest,dnu_rest,nuLnu_rest,dnuLnu_rest
            
                
        outfile.close()

if __name__ == "__main__":
    main()



#! /usr/bin/env python27

import numpy as np
from scipy.optimize import leastsq
from .leastsqbound import leastsqbound

def func(p,x):
    """model data as y = m*x+b """
    m,b = p
    return m*np.array(x)+b

def err(p,y,x):
    return y-func(p,x)

# extract data
temp = np.genfromtxt("sample_data.dat")
x = temp[:,0]
y = temp[:,1]

# perform unbounded least squares fitting
p0 = [1.0,0.0]
p, cov_x, infodic, mesg, ier = leastsq(err, p0, args=(y, x), full_output=True)

# print out results 
# print "Standard Least Squares fitting results:"
# print "p:", p
# print "cov_x:", cov_x
# print "infodic['nfev']:", infodic['nfev']
# print "infodic['fvec']:", infodic['fvec']
# print "infodic['fjac']:", infodic['fjac']
# print "infodic['ipvt']:", infodic['ipvt']
# print "infodic['qtf']:", infodic['qtf']
# print "mesg:", mesg
# print "ier:", ier
# print ""

# same as above using no bounds
p0 = [1.0,0.0]
p, cov_x, infodic, mesg, ier = leastsqbound(err, p0, args=(y, x), 
                                                full_output=True)

# print out results 
# print "Bounded Least Squares fitting with no bounds results:"
# print "p:", p
# print "cov_x:", cov_x
# print "infodic['nfev']:", infodic['nfev']
# print "infodic['fvec']:", infodic['fvec']
# print "infodic['fjac']:", infodic['fjac']
# print "infodic['ipvt']:", infodic['ipvt']
# print "infodic['qtf']:", infodic['qtf']
# print "mesg:", mesg
# print "ier:", ier
# print ""


# perform bounded least squares fitting
p0 = [1.0,0.0]
bounds = [(0.0, 2.0), (-10.0, 10.0)] 
p, cov_x, infodic, mesg, ier = leastsqbound(err, p0, args=(y, x),
                                            bounds = bounds, full_output=True)

# print out results 
# print "Bounded Least Squares fitting results:"
# print "p:", p
# print "cov_x:", cov_x
# print "infodic['nfev']:", infodic['nfev']
# print "infodic['fvec']:", infodic['fvec']
# print "infodic['fjac']:", infodic['fjac']
# print "infodic['ipvt']:", infodic['ipvt']
# print "infodic['qtf']:", infodic['qtf']
# print "mesg:", mesg
# print "ier:", ier



import sys
import pytest
from .base_class import TestBase

from .test_mcmc import TestEmcee
from .test_jet_model import TestJets
from .test_phenom_constr import TestPhenomenologyConstr
from .test_model_fit import TestModelFit
from .test_emitters import TestEmitters
from .test_ebl import TestEBL
from .test_mcmc import TestEmcee
from .test_depending_parameters import TestDependingParameters
from .test_composite_model import TestCompositeModel
from .test_temp_ev import TestTempEv

@pytest.fixture
def plot():
   input = False
   return input


class TestUser(TestBase):
  
   def test_jet(self,plot=plot):
      t=TestJets()
      t.integration_suite()

   def test_emitters(self,plot=plot):
      t=TestEmitters()
      t.integration_suite()

   def test_dep_pars(self,plot=plot):
      t=TestDependingParameters()
      t.integration_suite()
   
   def test_composit_model(self,plot=plot):
      t=TestCompositeModel()
      t.integration_suite()

   def test_composit_ebl(self,plot=plot):
      t=TestEBL()
      t.integration_suite()

      
   
  
  


__author__ = "Andrea Tramacere"

from .minimizer import  _eval_res

import emcee
from itertools import cycle


import numpy as np
import scipy as sp
from scipy import stats
import corner
import dill as pickle
from multiprocessing import cpu_count, Pool
import multiprocessing as mp
import warnings
import  time
import copy
import threading
from .plot_sedfit import  plt, PlotSED, set_mpl

__all__=['McmcSampler']



class Counter(object):


    def __init__(self,count_tot):
        self.count = 0
        self.count_OK = 0
        self.count_tot = count_tot
        self._progress_iter = cycle(['|', '/', '-', '\\'])


class RunThread(object):
    def __init__(self, target_class):
        self.target_class = target_class

    def run(self):
        self.target_class.model=self.target_class.model.clone()
        self.target_class.sampler.run_mcmc(self.target_class._pos, self.target_class._npernode, rstate0=np.random.get_state(), progress=True,store = True)


class McmcSampler(object):

    def __init__(self,model_minimizer):

        self.model = model_minimizer.fit_model
        self.data = model_minimizer.data
        self.fit_par_free = model_minimizer.fit_par_free

        self._progress_iter = cycle(['|', '/', '-', '\\'])



    def run_sampler(self,
                    nwalkers=500,
                    steps=100,
                    pos=None,
                    burnin=50,
                    use_UL=False,
                    bound=0.2,
                    bound_rel=False,
                    threads=None,
                    walker_start_bound=0.002,
                    use_labels_dict=None,
                    loglog = False,
                    progress='notebook'):

        counter=Counter(nwalkers*steps)
        self.calls = 0
        self.calls_OK = 0
        self.use_UL = use_UL

        self.pos = pos
        self.nwalkers = nwalkers
        self.steps = steps
        self.calls_tot = nwalkers * steps

        if use_labels_dict is None:
            self.par_array =  self.fit_par_free

            self.labels = [par.name for par in self.par_array]
            self.labels_units = [par.units for par in self.par_array]
            self.labels_start_val = [p.best_fit_val for p in self.par_array]
        else:
            self.labels=[]
            self.par_array=[]
            self.labels_units=[]
            self.labels_start_val=[]
            for model_name in use_labels_dict.keys():
                for par_name in use_labels_dict[model_name]:
                    p= self.model.parameters.get_par_by_name(model_name,par_name)
                    if p is not None:
                        self.par_array.append(p)
                        self.labels.append( p.name)
                        self.labels_units.append(p.units)
                        self.labels_start_val.append(p.best_fit_val)
                    else:
                        warnings.warn('par %s'%par_name+' not present in model, will be sckipped')

        self.par_array_best_fit=copy.deepcopy(self.par_array)
        self.ndim = len(self.labels)
        if pos is None:
            pos = emcee.utils.sample_ball(np.array([p.best_fit_val for p in self.par_array]),
                                          np.array([p.best_fit_val * walker_start_bound for p in self.par_array]),
                                          nwalkers)

        self._build_bounds(bound=bound,bound_rel=bound_rel)
        if emcee.__version__ < "3":
            raise RuntimeError('Please update to emcee v>=3.0.0')

        print('mcmc run starting')
        print('')
        start = time.time()
        if progress == 'notebook':
            pass

        if progress is True:
            tqdm=None


        if threads is not None and threads>1:
            # if threads > cpu_count():
            #     threads = cpu_count()
            #     warnings.warn('number of threads has been reduced to cpu_count=%d'%cpu_count())
            #


            # with Pool(processes=threads) as pool:
            #
            #     self.sampler = emcee.EnsembleSampler(nwalkers, self.ndim, log_prob, threads=threads,
            #                                          args=(self.model, self.data, use_UL, counter,self._bounds, self.par_array, loglog),pool=pool)
            #     self.sampler.run_mcmc(pos, steps, progress=True)



            warnings.warn('multithreadign not implemented yet')
            threads=1
            self.sampler = emcee.EnsembleSampler(nwalkers, self.ndim, log_prob, args=(self.model, self.data, use_UL, counter, self._bounds, self.par_array, loglog))
            self.sampler.run_mcmc(pos, steps, progress=progress)
        else:
            threads=1
            self.sampler = emcee.EnsembleSampler(nwalkers, self.ndim, log_prob,args=(self.model, self.data, use_UL, counter,self._bounds, self.par_array, loglog))
            self.sampler.run_mcmc(pos, steps, progress=progress)



        end = time.time()
        comp_time = end - start
        print("mcmc run done, with %d threads took %2.2f seconds"%(threads,comp_time))

        #self.samples = self.sampler.chain[:, burnin:, :].reshape((-1, self.ndim))
        self.samples = self.sampler.get_chain(flat=True,discard=burnin)
        self.samples_log_prob  = self.sampler.get_log_prob(flat=True,discard=burnin)
        self.acceptance_fraction=np.mean(self.sampler.acceptance_fraction)

        self.reset_to_best_fit()



    def get_par_quantiles(self,p,quantiles=(0.16,0.5,0.84)):
        _d, idx=self.get_par(p)

        return np.array(np.quantile(_d,quantiles))

    def reset_to_best_fit(self):
        for ID,par in enumerate(self.par_array):
            par.val = self.par_array_best_fit[ID].val

    def _build_bounds(self, bound=0.2,bound_rel=True):

        self._bounds=[]

        if np.shape(bound) == ():
            bound=[bound,bound]
        elif np.shape(bound) == (2,):
            pass
        else:
            raise RuntimeError('bound shape', np.shape(bound), 'it is wrong, has to be scalare or (2,)')

        for par in self.par_array:
            if  bound_rel is False and par.best_fit_err is not None:

                #_min =  par.best_fit_val - par.best_fit_err * bound[0]
                #_max =  par.best_fit_val + par.best_fit_err * bound[1]
                delta_p = par.best_fit_err * bound[1]
                delta_m = par.best_fit_err * bound[0]

            else:
                delta_p = np.fabs(par.best_fit_val)*bound[1]
                delta_m = np.fabs(par.best_fit_val)*bound[0]
                #_min = par.best_fit_val * (1.0  - bound[0])
                #_max = par.best_fit_val * (1.0  + bound[1])

            _min = par.best_fit_val - delta_m
            _max = par.best_fit_val + delta_p


            if par.fit_range_min is not None:
                _min= max(_min, par.fit_range_min)

            if par.fit_range_max is not None:
                _max= min(_max, par.fit_range_max)

            self._bounds.append([_min, _max])


    def show_pars(self):
        pass

    def corner_plot(self, labels = None, quantiles = (0.16, 0.5, 0.84), levels = None, title_kwargs = {}):
        _id = []

        if labels is None:
            labels=self.labels

        if type(labels) == list:
            pass
        else:
            labels = [labels]

        for l in labels:
            _id.append(self.labels.index(l))

        f = corner.corner(self.samples[:, _id],
                          quantiles=quantiles, labels=[self.labels[i] for i in _id],
                          truths=[self.labels_start_val[i] for i in _id],
                          title_kwargs=title_kwargs,show_titles = True,
                          levels = levels)

        title = 'quantiles ='+str(quantiles)
        f.suptitle(title,y=1.0)
        return f

    def get_par(self, p):
        if type(p) == int:
            pass

        else:
            try:
                p = self.labels.index(p)
            except:
                raise RuntimeError('paramter p', p, 'not found')

        if p > len(self.labels):
            raise RuntimeError('label id larger then labels size')

        return self.samples[:, p].flatten(), p


    def plot_chain(self, p, log_plot=False):
        _d, idx = self.get_par(p)

        n = self.labels[idx]

        traces=self.sampler.chain[:, :, idx]

        if self.labels_units is not None:
            if self.labels_units[idx] is not None:
                n += ' (%s)' % self.labels_units[idx]

        alpha_true = np.median(_d)

        f = plt.figure()
        ax = f.add_subplot(111)

        if log_plot == True:
            n = 'log10(%s)'%n
            if np.any(_d <= 0):
                raise RuntimeWarning('negative values in p')
            else:
                traces = np.log10(traces)
                alpha_true = np.log10(alpha_true)
        for t in traces:
            ax.plot(t, '-', color='k', alpha=0.5)



        ax.axhline(alpha_true, color='blue')

        ax.set_ylabel(n)
        ax.set_xlabel('steps')
        return f

    def plot_par(self, p, nbins=20, log_plot=False,quantiles=(0.16,0.5,0.84),figsize=None):
        set_mpl()

        _d, idx = self.get_par(p)

        par_name = self.labels[idx]

        x_name = par_name
        if self.labels_units is not None:
            if self.labels_units[idx] is not None and str(self.labels_units[idx]).strip() != '' :
                x_name += ' (%s)' % self.labels_units[idx]

        f = plt.figure(figsize=figsize)
        ax = f.add_subplot(111)

        if log_plot == True:
            x_name = 'log10(%s)' % x_name
            if np.any(_d <= 0):
                raise RuntimeWarning('negative values in p')
            else:
                _d = np.log10(_d)

        q_vals=self.get_par_quantiles(p,quantiles=quantiles)

        q_diff = np.diff(q_vals)

        ax.hist(_d,
                bins=nbins,
                density=True,
                alpha=0.5,
                label=r'%s = $%.3e^{+%.3e}_{-%.3e} $'%(par_name,q_vals[0],q_diff[1],q_diff[0]))

        for q in q_vals:
            if log_plot is True:
                q=np.log10(q)
            ax.axvline(q,c='black',ls='--',lw=0.5)

        ax.set_xlabel(x_name)
        f.suptitle('quantiles = %s'%str(quantiles))
        ax.legend(loc='center left', bbox_to_anchor=(1.0, 0.5), ncol=1)
        return f

    def plot_model(self, sed_data=None, fit_range=None, size=100, frame='obs', density=False):

        if sed_data is None:
            sed_data=self.sed_data

        if fit_range is None:
            fit_range = [self.model.nu_min_fit, self.model.nu_max_fit]

        p = self.model._set_up_plot(None, sed_data, frame, density)

        self.reset_to_best_fit()
        self.model.eval(fill_SED=True)

        x, y = self.model.SED.get_model_points(log_log=False, frame=frame)
        #if density is True:
        #    y=y-x
        if size is None:
            size = len(self.samples)
            ID_mcmc = np.arange(size)
        else:
            size = min(len(self.samples), int(size))
            ID_mcmc = np.random.randint(len(self.samples), size=size)

        y = np.zeros((size,x.size))

        for ID,ID_rand in enumerate(ID_mcmc):

            for ID_par, pi in enumerate(self.par_array):
                pi.set(val=self.get_par(ID_par)[0][ID_rand])

            self.model.eval(fill_SED=True)
            x, y[ID] = self.model.SED.get_model_points(log_log=False, frame=frame)

        if density is True:
            y=y/x
        y_min=np.amin(y, axis=0)
        y_max=np.amax(y, axis=0)
        msk = y_min > np.log10(self.model.flux_plot_lim)
        p.sedplot.fill_between(x[msk],y_max[msk],y_min[msk],color='gray',alpha=0.3,label='mcmc model range')

        self.reset_to_best_fit()
        self.model.eval(fill_SED=True)

        p.add_model_plot(self.model, color='red',fit_range = fit_range,density=density,flim=self.model.flux_plot_lim)
        p.add_model_residual_plot(model = self.model, data = sed_data, fit_range =  fit_range, color='red')


        return p


    def _progess_bar(self,):
        if np.mod(self.calls, 10) == 0 and self.calls != 0:
            print("\r%s progress=%3.3f%% calls=%d accepted=%d" % (next(self._progress_iter),float(100*self.calls)/(self.calls_tot),self.calls,self.calls_OK), end="")

    def save(self, name):
        with open(name, 'wb') as output:
            pickle.dump(self, output, pickle.HIGHEST_PROTOCOL)

    @classmethod
    def load(self, name):
        with open(name, 'rb') as input:
            return pickle.load(input)



def emcee_log_like(theta,fit_model,data,use_UL,par_array,loglog):
    _warn = False
    for pi in range(len(theta)):
        par_array[pi].set(val=theta[pi])
        if np.isnan(theta[pi]):
            _warn=True

    _m = fit_model.eval(nu=data['x'], fill_SED=False, get_model=True, loglog=loglog)

    _res_sum, _res, _res_UL = _eval_res(data['y'],
                                        _m,
                                        data['dy'],
                                        data['UL'],
                                        use_UL=use_UL)
    #_progess_bar(counter)
    return  _res_sum *-0.5


def _progess_bar(counter):

    if np.mod(counter.count, 10) == 0 and counter.count != 0:
        print("\r%s progress=%3.3f%% calls=%d accepted=%d" % (next( counter._progress_iter),float(100* counter.count)/( counter.count_tot),counter.count,counter.count_OK), end="")



def log_prob(theta,fit_model,data,use_UL,counter,bounds,par_array,loglog):
    lp = log_prior(theta,bounds)
    counter.count += 1
    if not np.isfinite(lp):
        lp = -np.inf
    else:
        lp += emcee_log_like(theta,fit_model,data,use_UL,par_array,loglog)
    counter.count_OK += 1
    return lp




def log_prior(theta,bounds):
    _r=0.
    #bounds = [(par.fit_range_min, par.fit_range_max) for par in model_minimizer.fit_par_free]
    #skip=False
    for pi in range(len(theta)):

        if bounds[pi][1] is not None:
            if theta[pi]>bounds[pi][1]:
                _r=-np.inf

        if bounds[pi][0] is not None:
            if theta[pi]<bounds[pi][0]:
                _r=-np.inf


    #if skip is True:
    #_r=-np.inf
    #print(theta[pi],bounds)
    return _r





__author__ = "Andrea Tramacere"


from .model_parameters import ModelParameterArray, ModelParameter

import numpy as np
from astropy import constants as const
from .spectral_shapes import SED
from .base_model import  Model


__all__=['AnalyticalParameter','Disk']

class AnalyticalParameter(ModelParameter):
    """
    This class is a subclass of the :class:`.ModelParameter` class,
    extending the base class to analytical parameters.
    """
    def __init__(self,polymodel,**keywords):
        
        self.polymodel=polymodel

        self.allowed_par_types=['Temperature','peak freq','peak flux']

        super(AnalyticalParameter,self).__init__(  **keywords)

        if 'val' in keywords.keys():
            val=keywords['val']
            self.assign_val(self.name,val)
            
        
        
    def set(self,**keywords):
        super(AnalyticalParameter,self).set(**keywords )
        
        """
        sets a parameter value checking for physical boundaries 
        """
        
        if 'val' in keywords.keys():    
            self.assign_val(self.name,keywords['val']) 
        
    
    def assign_val(self,name,val):
        """
        assigns the paramter valut to the :class:`Template` object
        """
        setattr(self.polymodel,name,val)






class Disk(Model):
    def __init__(self,cosmo,z,nu_size=100,name='disk',**keywords):
        """
        """
        
        super(Disk,self).__init__(  **keywords)
        
        self.z=z

        self.name=name

        self.model_type='analytical'
            
        self.SED = SED(name=self.model_type)
    
        self.parameters = ModelParameterArray()
      
        self.T_Disk=10000
        self.parameters.add_par(AnalyticalParameter(self,name='T_Disk',par_type='Temperature',val=10000,val_min=0.,val_max=None,units=''))
        
       
        self.nuFnu_p_D=1E-10
        self.parameters.add_par(AnalyticalParameter(self,name='nuFnu_p',par_type='peak flux',val=1E-10,val_min=0.,val_max=1E-3,units='erg cm^-2 s^-1',log=True))
        
        self.HPLANCK=const.h.cgs.value
        self.SIGSB=const.sigma_sb.cgs.value
        self.vluce_cm=const.c.cgs.value
        self.K_boltz=const.k_B.cgs.value


        self.cosmo=cosmo

        self.DL=self.cosmo.get_DL_cm(self.z)
        
        self.L_D=self.get_L_D()
        self.nu_p_Disk=self.get_nu_p()
        
        

    def set_disk_pars(self,fit_model,model_name):
        
        self.L_Disk=self.get_L_D()
          
        self.L_Disk_err=self.L_Disk*(fit_model.parameters.get(model_name,'nuFnu_p','best_fit_err')/
                                     fit_model.parameters.get(model_name,'nuFnu_p','best_fit_val'))

                        
        self.T_Disk=fit_model.parameters.get(model_name,'T_Disk','best_fit_val')
        self.T_Disk_err=fit_model.parameters.get(model_name,'T_Disk','best_fit_err')

        self.nu_p_Disk=self.get_nu_p()
        self.nu_p_Disk_err=self.nu_p_Disk*(self.T_Disk_err/self.T_Disk)

    
    def get_nu_p(self):
        return self.T_Disk*(1.39*5.879e10)

    
    def get_L_D(self):
        return 4*np.pi*self.DL*self.DL*self.nuFnu_p
        
        
    #def log_func(self,log_nu):
    #    return np.log10(self.lin_func(np.power(10,log_nu)))

    
    def lin_func(self,nu):
        nu=nu*(1+self.z)
        a = 2 * self.HPLANCK * nu*nu*nu*nu / (self.vluce_cm* self.vluce_cm)
        a *= 1.0 / (np.exp((self.HPLANCK * nu) / (self.K_boltz * self.T_Disk)) - 1)
        return a/(self.SIGSB*self.T_Disk*self.T_Disk*self.T_Disk*self.T_Disk/np.pi)*self.nuFnu_p
    
    
    def log_func(self,log_nu):
        
        return np.log10(self.lin_func(np.power(10,log_nu)))


"""Constrained multivariate least-squares optimization"""


from __future__ import absolute_import, division, print_function,unicode_literals

from builtins import (bytes, str, open, super, range,
                      zip, round, input, int, pow, object, map, zip)

__author__ = "Andrea Tramacere"

import warnings

from numpy import array, take, eye, triu, transpose, dot
from numpy import empty_like, sqrt, cos, sin, arcsin
try:
    from scipy.optimize.minpack import _check_func
except:
    from numpy import (atleast_1d, dot, take, triu, shape, eye,
                   transpose, inexact, issubdtype, dtype)
    def _check_func(checker, argname, thefunc, x0, args, numinputs,
                output_shape=None):
        res = atleast_1d(thefunc(*((x0[:numinputs],) + args)))
        if (output_shape is not None) and (shape(res) != output_shape):
            if (output_shape[0] != 1):
                if len(output_shape) > 1:
                    if output_shape[1] == 1:
                        return shape(res)
                msg = "%s: there is a mismatch between the input and output " \
                    "shape of the '%s' argument" % (checker, argname)
                func_name = getattr(thefunc, '__name__', None)
                if func_name:
                    msg += " '%s'." % func_name
                else:
                    msg += "."
                msg += 'Shape should be %s but it is %s.' % (output_shape, shape(res))
                raise TypeError(msg)
        if issubdtype(res.dtype, inexact):
            dt = res.dtype
        else:
            dt = dtype(float)
        return shape(res), dt 
        
from scipy.optimize import _minpack, leastsq


def _internal2external_grad(xi, bounds):
    """
    Calculate the internal (unconstrained) to external (constained)
    parameter gradiants.
    """
    grad = empty_like(xi)
    for i, (v, bound) in enumerate(zip(xi, bounds)):
        lower, upper = bound
        if lower is None and upper is None:  # No constraints
            grad[i] = 1.0
        elif upper is None:     # only lower bound
            grad[i] = v / sqrt(v * v + 1.)
        elif lower is None:     # only upper bound
            grad[i] = -v / sqrt(v * v + 1.)
        else:   # lower and upper bounds
            grad[i] = (upper - lower) * cos(v) / 2.
    return grad


def _internal2external_func(bounds):
    """
    Make a function which converts between internal (unconstrained) and
    external (constrained) parameters.
    """
    ls = [_internal2external_lambda(b) for b in bounds]

    def convert_i2e(xi):
        xe = empty_like(xi)
        xe[:] = [l(p) for l, p in zip(ls, xi)]
        return xe

    return convert_i2e


def _internal2external_lambda(bound):
    """
    Make a lambda function which converts a single internal (uncontrained)
    parameter to a external (constrained) parameter.
    """
    lower, upper = bound

    if lower is None and upper is None:  # no constraints
        return lambda x: x
    elif upper is None:     # only lower bound
        return lambda x: lower - 1. + sqrt(x * x + 1.)
    elif lower is None:     # only upper bound
        return lambda x: upper + 1. - sqrt(x * x + 1.)
    else:
        return lambda x: lower + ((upper - lower) / 2.) * (sin(x) + 1.)


def _external2internal_func(bounds):
    """
    Make a function which converts between external (constrained) and
    internal (unconstrained) parameters.
    """
    ls = [_external2internal_lambda(b) for b in bounds]

    def convert_e2i(xe):
        xi = empty_like(xe)
        xi[:] = [l(p) for l, p in zip(ls, xe)]
        return xi

    return convert_e2i


def _external2internal_lambda(bound):
    """
    Make a lambda function which converts an single external (constrained)
    parameter to a internal (unconstrained) parameter.
    """
    lower, upper = bound

    if lower is None and upper is None:  # no constraints
        return lambda x: x
    elif upper is None:     # only lower bound
        return lambda x: sqrt((x - lower + 1.) ** 2 - 1)
    elif lower is None:     # only upper bound
        return lambda x: sqrt((upper - x + 1.) ** 2 - 1)
    else:
        return lambda x: arcsin((2. * (x - lower) / (upper - lower)) - 1.)


def leastsqbound(func, x0, args=(), bounds=None, Dfun=None, full_output=0,
            col_deriv=0, ftol=1.49012e-8, xtol=1.49012e-8,
            gtol=0.0, maxfev=0, epsfcn=0.0, factor=100, diag=None):
    """
    Bounded minimization of the sum of squares of a set of equations.

    ::

        x = arg min(sum(func(y)**2,axis=0))
                 y

    Parameters
    ----------
    func : callable
        should take at least one (possibly length N vector) argument and
        returns M floating point numbers.
    x0 : ndarray
        The starting estimate for the minimization.
    args : tuple
        Any extra arguments to func are placed in this tuple.
    bounds : list
        ``(min, max)`` pairs for each element in ``x``, defining
        the bounds on that parameter. Use None for one of ``min`` or
        ``max`` when there is no bound in that direction.
    Dfun : callable
        A function or method to compute the Jacobian of func with derivatives
        across the rows. If this is None, the Jacobian will be estimated.
    full_output : bool
        non-zero to return all optional outputs.
    col_deriv : bool
        non-zero to specify that the Jacobian function computes derivatives
        down the columns (faster, because there is no transpose operation).
    ftol : float
        Relative error desired in the sum of squares.
    xtol : float
        Relative error desired in the approximate solution.
    gtol : float
        Orthogonality desired between the function vector and the columns of
        the Jacobian.
    maxfev : int
        The maximum number of calls to the function. If zero, then 100*(N+1) is
        the maximum where N is the number of elements in x0.
    epsfcn : float
        A suitable step length for the forward-difference approximation of the
        Jacobian (for Dfun=None). If epsfcn is less than the machine precision,
        it is assumed that the relative errors in the functions are of the
        order of the machine precision.
    factor : float
        A parameter determining the initial step bound
        (``factor * || diag * x||``). Should be in interval ``(0.1, 100)``.
    diag : sequence
        N positive entries that serve as a scale factors for the variables.

    Returns
    -------
    x : ndarray
        The solution (or the result of the last iteration for an unsuccessful
        call).
    cov_x : ndarray
        Uses the fjac and ipvt optional outputs to construct an
        estimate of the jacobian around the solution.  ``None`` if a
        singular matrix encountered (indicates very flat curvature in
        some direction).  This matrix must be multiplied by the
        residual standard deviation to get the covariance of the
        parameter estimates -- see curve_fit.
    infodict : dict
        a dictionary of optional outputs with the key s::

            - 'nfev' : the number of function calls
            - 'fvec' : the function evaluated at the output
            - 'fjac' : A permutation of the R matrix of a QR
                     factorization of the final approximate
                     Jacobian matrix, stored column wise.
                     Together with ipvt, the covariance of the
                     estimate can be approximated.
            - 'ipvt' : an integer array of length N which defines
                     a permutation matrix, p, such that
                     fjac*p = q*r, where r is upper triangular
                     with diagonal elements of nonincreasing
                     magnitude. Column j of p is column ipvt(j)
                     of the identity matrix.
            - 'qtf'  : the vector (transpose(q) * fvec).

    mesg : str
        A string message giving information about the cause of failure.
    ier : int
        An integer flag.  If it is equal to 1, 2, 3 or 4, the solution was
        found.  Otherwise, the solution was not found. In either case, the
        optional output variable 'mesg' gives more information.

    Notes
    -----
    "leastsq" is a wrapper around MINPACK's lmdif and lmder algorithms.

    cov_x is a Jacobian approximation to the Hessian of the least squares
    objective function.
    This approximation assumes that the objective function is based on the
    difference between some observed target data (ydata) and a (non-linear)
    function of the parameters `f(xdata, params)` ::

           func(params) = ydata - f(xdata, params)

    so that the objective function is ::

           min   sum((ydata - f(xdata, params))**2, axis=0)
         params

    Contraints on the parameters are enforced using an internal parameter list
    with appropiate transformations such that these internal parameters can be
    optimized without constraints. The transfomation between a given internal
    parameter, p_i, and a external parameter, p_e, are as follows:

    With ``min`` and ``max`` bounds defined ::

        p_i = arcsin((2 * (p_e - min) / (max - min)) - 1.)
        p_e = min + ((max - min) / 2.) * (sin(p_i) + 1.)

    With only ``max`` defined ::

        p_i = sqrt((max - p_e + 1.)**2 - 1.)
        p_e = max + 1. - sqrt(p_i**2 + 1.)

    With only ``min`` defined ::

        p_i = sqrt((p_e - min + 1.)**2 - 1.)
        p_e = min - 1. + sqrt(p_i**2 + 1.)

    These transfomations are used in the MINUIT package, and described in
    detail in the section 1.3.1 of the MINUIT User's Guide.

    To Do
    -----
    Currently the ``factor`` and ``diag`` parameters scale the
    internal parameter list, but should scale the external parameter list.

    The `qtf` vector in the infodic dictionary reflects internal parameter
    list, it should be correct to reflect the external parameter list.

    References
    ----------
    * F. James and M. Winkler. MINUIT User's Guide, July 16, 2004.

    """
    # use leastsq if no bounds are present
    if bounds is None:
        return leastsq(func, x0, args, Dfun, full_output, col_deriv,
                        ftol, xtol, gtol, maxfev, epsfcn, factor, diag)

    # create function which convert between internal and external parameters
    i2e = _internal2external_func(bounds)
    e2i = _external2internal_func(bounds)

    x0 = array(x0, ndmin=1)
    i0 = e2i(x0)
    n = len(x0)
    if len(bounds) != n:
        raise ValueError('length of x0 != length of bounds')
    if type(args) != type(()):
        args = (args,)

    m = _check_func('leastsq', 'func', func, x0, args, n)[0][0]
    #print ('m',m)
    if n > m:
        raise TypeError('Improper input: N=%s must not exceed M=%s' % (n, m))

    # define a wrapped func which accept internal parameters, converts them
    # to external parameters and calls func
    def wfunc(x, *args):
        return func(i2e(x), *args)

    if Dfun is None:
        if (maxfev == 0):
            maxfev = 200 * (n + 1)
        retval = _minpack._lmdif(wfunc, i0, args, full_output, ftol, xtol,
                gtol, maxfev, epsfcn, factor, diag)
    else:
        if col_deriv:
            _check_func('leastsq', 'Dfun', Dfun, x0, args, n, (n, m))
        else:
            _check_func('leastsq', 'Dfun', Dfun, x0, args, n, (m, n))
        if (maxfev == 0):
            maxfev = 100 * (n + 1)

        def wDfun(x, *args):  # wrapped Dfun
            return Dfun(i2e(x), *args)

        retval = _minpack._lmder(func, wDfun, i0, args, full_output,
                col_deriv, ftol, xtol, gtol, maxfev, factor, diag)

    errors = {0: ["Improper input parameters.", TypeError],
              1: ["Both actual and predicted relative reductions "
                 "in the sum of squares\n  are at most %f" % ftol, None],
              2: ["The relative error between two consecutive "
                 "iterates is at most %f" % xtol, None],
              3: ["Both actual and predicted relative reductions in "
                 "the sum of squares\n  are at most %f and the "
                 "relative error between two consecutive "
                 "iterates is at \n  most %f" % (ftol, xtol), None],
              4: ["The cosine of the angle between func(x) and any "
                 "column of the\n  Jacobian is at most %f in "
                 "absolute value" % gtol, None],
              5: ["Number of calls to function has reached "
                 "maxfev = %d." % maxfev, ValueError],
              6: ["ftol=%f is too small, no further reduction "
                 "in the sum of squares\n  is possible.""" % ftol, ValueError],
              7: ["xtol=%f is too small, no further improvement in "
                 "the approximate\n  solution is possible." % xtol, ValueError],
              8: ["gtol=%f is too small, func(x) is orthogonal to the "
                 "columns of\n  the Jacobian to machine "
                 "precision." % gtol, ValueError],
              'unknown': ["Unknown error.", TypeError]}

    info = retval[-1]    # The FORTRAN return value

    if (info not in [1, 2, 3, 4] and not full_output):
        if info in [5, 6, 7, 8]:
            warnings.warn(errors[info][0], RuntimeWarning)
        else:
            try:
                raise errors[info][1](errors[info][0])
            except KeyError:
                raise errors['unknown'][1](errors['unknown'][0])

    mesg = errors[info][0]
    x = i2e(retval[0])  # internal params to external params

    if full_output:
        # convert fjac from internal params to external
        grad = _internal2external_grad(retval[0], bounds)
        retval[1]['fjac'] = (retval[1]['fjac'].T / take(grad,
                            retval[1]['ipvt'] - 1)).T
        cov_x = None
        if info in [1, 2, 3, 4]:
            from numpy.dual import inv
            from numpy.linalg import LinAlgError
            perm = take(eye(n), retval[1]['ipvt'] - 1, 0)
            r = triu(transpose(retval[1]['fjac'])[:n, :])
            R = dot(r, perm)
            try:
                cov_x = inv(dot(transpose(R), R))
            except LinAlgError:
                pass
        return (x, cov_x) + retval[1:-1] + (mesg, info)
    else:
        return (x, info)



__author__ = "Andrea Tramacere"


import json
import dill as pickle
import six
import numpy as np
import copy
import warnings
import os

from .jet_spectral_components import JetSpecComponent, SpecCompList

from .model_parameters import ModelParameterArray, ModelParameter, _show_table
from .base_model import Model
from .output import makedir,WorkPlace
from  .plot_sedfit import PlotSED,plt
from .cosmo_tools import Cosmo
from .utils import safe_run,set_str_attr, old_model_warning, get_info, clean_var_name
from .jet_paramters import *
from .jet_emitters import *
from .jet_emitters_factory import EmittersFactory
from .jet_tools import *
from .mathkernel_helper import bessel_table_file_path


on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

if on_rtd is True:
    try:
        from .jetkernel import jetkernel as BlazarSED
    except ImportError:
        from .mock import jetkernel as BlazarSED
else:
    from .jetkernel import jetkernel as BlazarSED



__all__=['Jet','JetBase']


class JetBase(Model):
    """ JetBase class.
    This class allows to build a ``Jet`` model providing the interface to the
    C code, giving  full access to the physical parameters and
    providing the methods to run the code.
    A :class:`Jet` object  will store the
    the physical parameters in  the ::py:attr:`Jet.parameters`  that is :class:`.ModelParameterArray` class,
    i.e. a collection of :class:`JetParameter` objects.
    All the physical parameters are  also accessible as attributes of
    the  ::py:attr:`Jet.parameters`
    """
    def __repr__(self):
        return str(self.show_model())

    

    def __init__(self,
                 cosmo=None,
                 name='test',
                 emitters_type='electrons',
                 emitters_distribution='pl',
                 emitters_distribution_log_values=False,
                 beaming_expr='delta',
                 jet_workplace=None,
                 verbose=None,
                 nu_size=500,
                 clean_work_dir=True,
                 **keywords):

        """

        Parameters
        ----------
        cosmo
        name
        emitters_type
        emitters_distribution
        emitters_distribution_log_values
        beaming_expr
        jet_workplace
        verbose
        nu_size
        clean_work_dir
        """
        super(JetBase,self).__init__(  **keywords)




        if cosmo is not None:
            self.cosmo=cosmo
        else:
            self.cosmo= Cosmo()
        #print('cosmo', self.cosmo)
        self.name = clean_var_name(name)

        self.model_type='jet'
        self._emitters_type=emitters_type
        self._scale='lin-lin'

        self._blob = self.build_blob(verbose=verbose)
        self._static_spec_arr_grid_size = BlazarSED.static_spec_arr_grid_size
        self._nu_static_size = BlazarSED.static_spec_arr_size
        self.nu_size = nu_size
        self.nu_grid_size=self._get_nu_grid_size_blob()
        if jet_workplace is None:
            jet_workplace=WorkPlace()
            out_dir= jet_workplace.out_dir + '/' + self.name + '_jet_prod/'
        else:
            out_dir=jet_workplace.out_dir+'/'+self.name+'_jet_prod/'

        self.set_path(out_dir,clean_work_dir=clean_work_dir)

        self.set_flag(self.name)


        self._allowed_EC_components_list=['EC_BLR',
                                          'DT',
                                          'EC_DT',
                                          'Star',
                                          'EC_Star',
                                          #CMB',
                                          'EC_CMB',
                                          'Disk',
                                          'Disk_MultiBB',
                                          'Disk_Mono',
                                          'EC_Disk',
                                          'All']

        self._allwed_disk_type =['BB', 'Mono', 'MultiBB']
        self.EC_components_list =[]
        self.spectral_components_list=[]


        self.spectral_components= SpecCompList(self.spectral_components_list)

        self.add_basic_components()
        self.SED=self.get_spectral_component_by_name('Sum').SED

        self.parameters = JetModelParameterArray(model=self)

        self._emitting_region_dic = None
        self._electron_distribution_dic= None
        self._external_photon_fields_dic= None
        self._original_emitters_distr = None

        self._setup(emitters_distribution,emitters_distribution_log_values,beaming_expr,emitters_type)


    def _setup(self, emitters_distribution, emitters_distribution_log_values, beaming_expr, emitters_type):
        self.EC_components_list = []
        self.spectral_components_list = []

        self.spectral_components = SpecCompList(self.spectral_components_list)
        self.add_basic_components()

        self.SED = self.get_spectral_component_by_name('Sum').SED

        self.parameters = JetModelParameterArray(model=self)

        self._emitting_region_dic = None
        self._emitters_distribution_dic = None
        self._external_photon_fields_dic = None



        self._blob.IC_adaptive_e_binning = 0
        self._blob.do_IC_down_scattering = 0

        self.set_emitting_region(beaming_expr)

        self.flux_plot_lim=1E-30
        self.set_emiss_lim(1E-120)

        self._IC_states = {}
        self._IC_states['on'] = 1
        self._IC_states['off'] = 0

        self._external_field_transf = {}
        self._external_field_transf['blob'] = 0
        self._external_field_transf['disk'] = 1
        self._jetkernel_interp='linear'
        self.set_emitters_distribution(emitters_distribution, emitters_distribution_log_values, emitters_type,
                                       init=False)
        self.set_blob()


    def __getstate__(self):
        return  self._serialize_model()

    def __setstate__(self,state):
        self.__init__()
        self._decode_model(state)

    def _serialize_model(self):

        _model = {}
        _model['version']=get_info()['version']
        _model['name'] = self.name
        if isinstance(self.emitters_distribution,JetkernelEmittersDistribution):
            _model['emitters_distribution'] = self._emitters_distribution_name
            _model['emitters_distribution_log_values'] = self._emitters_distribution_log_values
            _model['emitters_type'] = self._emitters_type
            _model['emitters_distribution_class']='JetkernelEmittersDistribution'
        elif isinstance(self.emitters_distribution,EmittersDistribution):
            self._original_emitters_distr._copy_from_jet(self)
            _model['custom_emitters_distribution'] =self._original_emitters_distr
            _model['emitters_distribution_class'] = 'EmittersDistribution'
        else:
            raise  RuntimeError('emitters distribuion type not valid',type(self._emitters_distribution))

        if hasattr(self,'T_esc_e_second'):
            _model['T_esc_e_second']=self.T_esc_e_second

        _model['beaming_expr'] = self._beaming_expr
        _model['spectral_components_name'] = self.get_spectral_component_names_list()
        _model['spectral_components_state'] = [c.state for c in self.spectral_components_list]
        _model['EC_components_name'] = self.EC_components_list
        _model['basic_components_name'] = self.basic_components_list
        _model['cosmo'] = self.cosmo

        _model['pars'] = {}
        _model['pars']=self.parameters._serialize_pars()

        _model['internal_pars'] = {}
        _model['internal_pars']['nu_size'] = self.nu_size
        _model['internal_pars']['nu_seed_size'] = self.nu_seed_size
        _model['internal_pars']['gamma_grid_size'] = self.gamma_grid_size
        _model['internal_pars']['IC_nu_size']=self.IC_nu_size
        _model['internal_pars']['nu_min']=self.nu_min
        _model['internal_pars']['nu_max']=self.nu_max
        _model['internal_pars']['Norm_distr'] = self.Norm_distr
        return _model

    def save_model(self,file_name):
        pickle.dump(self._serialize_model(), open(file_name, 'wb'), protocol=pickle.HIGHEST_PROTOCOL)

    @classmethod
    @safe_run
    def load_model(cls, file_name):
        try:
            _model = pickle.load(open(file_name, "rb"))
        except Exception as e:
            raise RuntimeError('The model you loaded is not valid please check the file name', e)

        try:
            jet = cls(name='no_name')
            jet._decode_model(_model)
            jet._fix_par_dep_on_load()
            jet.show_pars()
            jet.eval()
            return jet
        except Exception as e:
            if 'version' in _model.keys():
                v=_model['version']
            else:
                v='unknown(<1.1.2)'
            msg = "trying  to load a model saved with jetset version %s \n "%v
            msg += "currenlty using jetset version %s \n "%get_info()['version']
            msg += "caused the following problem:\n %s\n"%(repr(e))

            raise RuntimeError (msg)




    def _decode_model(self,_model):

        if 'version' in _model.keys():
            self._set_version(_model['version'])
        else:
            self._set_version(v='unknown(<1.1.2)')

        self.cosmo = _model['cosmo']
        self.model_type = 'jet'
        self.name = _model['name']
        self.set_blob()
        self.parameters = JetModelParameterArray(model=self)


        if 'emitters_type' in _model.keys():
            emitters_type=str(_model['emitters_type'])
        else:
            emitters_type = 'electrons'

        if 'electron_distribution' in _model.keys():
            _v=_model['electron_distribution']
            del(_model['electron_distribution'])
            _model['emitters_distribution']=_v

        if 'electron_distribution_log_values' in _model.keys():
            _v=_model['electron_distribution_log_values']
            del(_model['electron_distribution_log_values'])
            _model['emitters_distribution_log_values']=_v

        if _model['emitters_distribution_class'] == 'JetkernelEmittersDistribution':
            self.set_emitters_distribution(distr=_model['emitters_distribution'],
                                           log_values=_model['emitters_distribution_log_values'],
                                           emitters_type=emitters_type,
                                           init=False)
        elif _model['emitters_distribution_class'] == 'EmittersDistribution':
            self.set_emitters_distribution(distr=_model['custom_emitters_distribution'], init=False)
        else:
            raise RuntimeError('emitters distribuion type not valid', type(self._emitters_distribution))

        for c in self.basic_components_list:
            if c not in _model['basic_components_name']:
                self.del_spectral_component(c)

        if self.emitters_distribution.emitters_type == 'protons':

            self.add_pp_gamma_component()
            self.add_pp_neutrino_component()
            self.add_bremss_ep_component()
            self.T_esc_e_second=_model['T_esc_e_second']

        #for k in _model['pars'].keys():
            #print ('-->',k,_model['pars'][k])
        if 'disk_type' in _model['pars'].keys():
            disk_type=_model['pars']['disk_type']['val']
        else:
            disk_type=None
        self.add_EC_component(_model['EC_components_name'],disk_type=disk_type)

        for ID, c in enumerate(_model['spectral_components_name']):
            comp = getattr(self.spectral_components, c)
            if comp._state_dict != {}:
                comp.state = _model['spectral_components_state'][ID]

        self.SED = self.get_spectral_component_by_name('Sum').SED

        self.set_emitting_region(str(_model['beaming_expr']))
        #self.set_electron_distribution(str(_model['electron_distribution']))

        _par_dict = _model['pars']
        _non_user_dict={}
        _user_dict={}
        for k, v in _model['pars'].items():
            if v['par_type'] == 'user_defined':
                _user_dict[k]=v
            else:
                _non_user_dict[k]=v
        self.parameters._decode_pars(_non_user_dict)

        for k, v in _user_dict.items():
            #print('==>',v)
            v['name']=k
            self.parameters.add_par(ModelParameter(**v))

        _par_dict = _model['internal_pars']
        for k in _par_dict.keys():
            #print ('set', k,_par_dict[k])
            setattr(self,k,_par_dict[str(k)])
            #self.set_par(par_name=str(k), val=_par_dict[str(k)])

    @classmethod
    def load_old_model(cls, file_name):

        old_model_warning()
        jet = cls()
        with open(file_name, 'r') as infile:
            _model = json.load(infile)

        # print ('_model',_model)

        jet.model_type = 'jet'

        jet.init_BlazarSED()

        jet.parameters = JetModelParameterArray(model=jet)

        if 'electron_distribution' in _model.keys():
            _v = _model['electron_distribution']
            del (_model['electron_distribution'])
            _model['emitters_distribution'] = _v

        if 'electron_distribution_log_values' in _model.keys():
            _v = _model['electron_distribution_log_values']
            del (_model['electron_distribution_log_values'])
            _model['emitters_distribution_log_values'] = _v

        jet.set_emitters_distribution(name=str(_model['emitters_distribution']),
                                       log_values=_model['emitters_distribution_log_values'],
                                       emitters_type='electrons',
                                       init=False)

        for c in jet.basic_components_list:
            if c not in _model['basic_components_name']:
                jet.del_spectral_component(c)


        jet.add_EC_component(_model['EC_components_name'])

        for ID, c in enumerate(_model['spectral_components_name']):
            comp = getattr(jet.spectral_components, c)
            if comp._state_dict != {}:
                comp.state = _model['spectral_components_state'][ID]

        jet.SED = jet.get_spectral_component_by_name('Sum').SED

        jet.set_emitting_region(str(_model['beaming_expr']))

        _par_dict = _model['pars']
        jet.show_pars()
        for k in _par_dict.keys():
            # print ('set', k,_par_dict[k])
            jet.set_par(par_name=str(k), val=_par_dict[str(k)])

        jet.eval()
        return jet



    def build_blob(self, verbose=None):

        blob = BlazarSED.MakeBlob()

        blob.x_Bessel_min = 1E-17
        blob.x_Bessel_max = 7.2E2

        blob.x_ave_Bessel_min = 1E-16
        blob.x_ave_Bessel_max = 3.5E2

        blob.log_x_Bessel_min = np.log10( blob.x_Bessel_min)
        blob.log_x_Bessel_max = np.log10( blob.x_Bessel_max)


        blob.log_x_ave_Bessel_min = np.log10( blob.x_ave_Bessel_min)
        blob.log_x_ave_Bessel_max = np.log10( blob.x_ave_Bessel_max)




        F_Sync_x_ptr = getattr(blob, 'F_Sync_x')

        F_Sync_y_ptr = getattr(blob,  'F_Sync_y')

        F_ave_Sync_x_ptr = getattr(blob,  'F_ave_Sync_x')

        F_ave_Sync_y_ptr = getattr(blob, 'F_ave_Sync_y')

        log_F_Sync_x_ptr = getattr(blob, 'log_F_Sync_x')

        log_F_Sync_y_ptr = getattr(blob, 'log_F_Sync_y')

        log_F_ave_Sync_x_ptr = getattr(blob, 'log_F_ave_Sync_x')

        log_F_ave_Sync_y_ptr = getattr(blob, 'log_F_ave_Sync_y')


        d = np.genfromtxt(bessel_table_file_path)
        log_F_Sync_x=np.log10(d[:,0])
        log_F_Sync_y=np.log10(d[:,1])
        log_F_ave_Sync_x=np.log10(d[:,2])
        log_F_ave_Sync_y=np.log10(d[:,3])

        for ID, l in enumerate(d):
            BlazarSED.set_bessel_table(F_Sync_x_ptr,blob, l[0], ID)
            BlazarSED.set_bessel_table(F_Sync_y_ptr, blob,  l[1], ID)
            BlazarSED.set_bessel_table(F_ave_Sync_x_ptr, blob,  l[2], ID)
            BlazarSED.set_bessel_table(F_ave_Sync_y_ptr, blob,  l[3], ID)

            BlazarSED.set_bessel_table(log_F_Sync_x_ptr, blob, log_F_Sync_x[ID], ID)
            BlazarSED.set_bessel_table(log_F_Sync_y_ptr, blob, log_F_Sync_y[ID], ID)
            BlazarSED.set_bessel_table(log_F_ave_Sync_x_ptr, blob, log_F_ave_Sync_x[ID], ID)
            BlazarSED.set_bessel_table(log_F_ave_Sync_y_ptr, blob, log_F_ave_Sync_y[ID], ID)

        blob.BESSEL_TABLE_DONE=1

        if verbose is None:
            blob.verbose = 0
        else:
            blob.verbose = verbose

        set_str_attr(blob, 'path', './')

        set_str_attr(blob, 'MODE', 'custom')

        blob.gamma_grid_size = 200

        blob.nu_IC_size = 100
        blob.nu_seed_size = 100

        blob.nu_grid_size= 1000

        blob.do_Sync = 2

        blob.do_SSC = 1

        blob.R = 5.0e15

        blob.B = 0.1

        blob.z_cosm = 0.1

        blob.BulkFactor = 10
        blob.theta = 0.1

        blob.N = 100


        blob.NH_pp = 1

        blob.L_Disk = 1E45

        blob.L_DT = 1E45

        blob.gmin = 2

        blob.gmax = 1e6

        blob.nu_start_Sync = 1e6
        blob.nu_stop_Sync = 1e20

        blob.nu_start_SSC = 1e14
        blob.nu_stop_SSC = 1e30

        blob.nu_start_grid = 1e6
        blob.nu_stop_grid = 1e30

        return blob

    def set_emitting_region(self,beaming_expr):

        if  self._emitting_region_dic is not None:
            self.del_par_from_dic(self._emitting_region_dic)

        self._beaming_expr=beaming_expr

        set_str_attr(self._blob,'BEAMING_EXPR',beaming_expr)

        self._emitting_region_dic=build_emitting_region_dic(self.cosmo,beaming_expr=beaming_expr)

        self.parameters.add_par_from_dict(self._emitting_region_dic,self,'_blob',JetParameter)

    @property
    def IC_adaptive_e_binning(self,):
        return np.int(self._blob.IC_adaptive_e_binning)

    @IC_adaptive_e_binning.setter
    def IC_adaptive_e_binning(self,state):
        if type(state) == bool:
            pass
        else:
            raise RuntimeError('state has to be boolean')
        self._blob.IC_adaptive_e_binning=np.int(state)

    @staticmethod
    def available_emitters_distributions():
        EmittersFactory.available_distributions()

    def set_emitters_distribution(self, distr=None, log_values=False, emitters_type='electrons', init=True):

        if init is True:
            self.set_blob()
        self._emitters_distribution_log_values = log_values

        if self._emitters_distribution_dic is not None:
            self.del_par_from_dic(self._emitters_distribution_dic)

        if isinstance(distr, ArrayDistribution):
            self._emitters_distribution_name = 'from_array'
            self.emitters_distribution = JetkernelEmittersDistribution.from_array(self, distr, emitters_type=emitters_type)

        elif isinstance(distr, JetkernelEmittersDistribution):
            self.emitters_distribution = distr

            self._emitters_distribution_name = self.emitters_distribution.name
            self._emitters_distribution_dic = self.emitters_distribution._parameters_dict

            self.parameters.add_par_from_dict(self._emitters_distribution_dic,self,'_blob',JetParameter)


        elif isinstance(distr, EmittersDistribution):
            self._original_emitters_distr = distr
            self.emitters_distribution = copy.deepcopy(distr)
            self._update_emitters_pars_dependence()
            self.emitters_distribution.set_jet(self)
            self.emitters_distribution._update_parameters_dict()
            self._emitters_distribution_name = self.emitters_distribution.name
            self._emitters_distribution_dic = self.emitters_distribution._parameters_dict
            self.parameters.add_par_from_dict(self._emitters_distribution_dic, self, '_blob', JetParameter)
            self._attach_pars_to_jet(preserve_value_emitters=True)

            self.emitters_distribution.update()
        elif isinstance(distr, str):
            nf=EmittersFactory()
            self.emitters_distribution = nf.create_emitters(distr, log_values=log_values, emitters_type=emitters_type)
            self._original_emitters_distr = copy.deepcopy(self.emitters_distribution)
            self.emitters_distribution.set_jet(self)
            self.emitters_distribution._update_parameters_dict()
            self._emitters_distribution_name = self.emitters_distribution.name
            self._emitters_distribution_dic = self.emitters_distribution._parameters_dict
            self.parameters.add_par_from_dict(self._emitters_distribution_dic, self, '_blob', JetParameter)
            self._attach_pars_to_jet(preserve_value_emitters=True)
            self._update_emitters_pars_dependence()
            self.emitters_distribution.update()
        else:
            raise RuntimeError('distr',type(distr),'not valid should be a string or an',type(EmittersDistribution),'instance')

    def _attach_pars_to_jet(self, preserve_value_emitters=False):

        v_dict={}
        for par in self.emitters_distribution.parameters.par_array[::]:
            self.emitters_distribution.parameters.del_par(par)
            v_dict[par.name]=par.val
        for k in self._emitters_distribution_dic.keys():
            par = self.parameters.get_par_by_name(k)
            if preserve_value_emitters is True:
                par.set(val=v_dict[k],skip_dep_par_warning=True)
            self.emitters_distribution.parameters.add_par(par)


    def _update_emitters_pars_dependence(self):
        for par in self.emitters_distribution.parameters.par_array:
            if par.immutable is True:
                warnings.warn('parameter dependence has to be reassigned after emitters distribution is assigned to a jet, now will be reset no dep')
                self.emitters_distribution.parameters.reset_dependencies()
                break

    def get_emitters_distribution_name(self):
        return self.emitters_distribution.name


    def show_spectral_components(self):

        print ("Spectral components for Jet model:%s"%(self.name))

        for comp in self.spectral_components_list:

            print ("comp: %s "%(comp.name))

        print()


    def get_spectral_component_by_name(self,name,verbose=True):
        for i in range(len(self.spectral_components_list)):
            if self.spectral_components_list[i].name==name:
                return self.spectral_components_list[i]
        else:
            if verbose==True:
                print ("no spectral components with name %s found"%name)
                print ("names in array are:")
                self.show_spectral_components()

            return None

    def list_spectral_components(self):
        for i in range(len(self.spectral_components_list)):
            print (self.spectral_components_list[i].name)

    def get_spectral_component_names_list(self):
        _l=[]
        for i in range(len(self.spectral_components_list)):
            _l.append(self.spectral_components_list[i].name)
        return _l

    def del_spectral_component(self,name):
        print('deleting spectral component', name)
        if name in self.EC_components_list:
            self.del_EC_component(name)
        else:
            self._del_spectral_component(name)
        if name in self.basic_components_list:
            self.basic_components_list.remove(name)

    def _del_spectral_component(self, name, verbose=True):

        comp=self.get_spectral_component_by_name(name,verbose=verbose)

        if comp._state_dict!={}:

            comp.state='off'

        if comp is not None:

            self.spectral_components_list.remove(comp)

            self.spectral_components.__delattr__(name)



    def _add_spectral_component(self, name, var_name=None, state_dict=None,state=None):

        self.spectral_components_list.append(
            JetSpecComponent(self, name, self._blob, var_name=var_name, state_dict=state_dict, state=state))
        setattr(self.spectral_components,name,self.spectral_components_list[-1])

    def _update_spectral_components(self):
        _l=[]
        for ID,s in enumerate(self.spectral_components_list):
            self.spectral_components_list[ID]= JetSpecComponent(self, s.name, self._blob, var_name=s._var_name, state_dict=s._state_dict, state=s.state)
            setattr(self.spectral_components, self.spectral_components_list[ID].name, self.spectral_components_list[ID])


    def add_basic_components(self):
        self.basic_components_list=['Sum','Sync','SSC']

        self._add_spectral_component('Sum')
        self._add_spectral_component('Sync', var_name='do_Sync', state_dict=dict((('on', 1), ('off', 0), ('self-abs', 2))),state='self-abs')

        self._add_spectral_component('SSC', var_name='do_SSC', state_dict=dict((('on', 1), ('off', 0))))


    def add_sync_component(self,state='self-abs'):
        self._add_spectral_component('Sync', var_name='do_Sync',
                                     state_dict=dict((('on', 1), ('off', 0), ('self-abs', 2))), state=state)

    def add_SSC_component(self,state='on'):
        self._add_spectral_component('SSC', var_name='do_SSC', state_dict=dict((('on', 1), ('off', 0))),state=state)

    def del_EC_component(self,EC_components_list, disk_type='BB'):
        if isinstance(EC_components_list, six.string_types):
            EC_components_list = [EC_components_list]

        if 'All' in EC_components_list:
            EC_components_list=self._allowed_EC_components_list[::]


        for EC_component in EC_components_list:


            if EC_component not in self._allowed_EC_components_list:
                raise RuntimeError("EC_component %s not allowed" % EC_component, "please choose among ",
                                   self._allowed_EC_components_list)

            if EC_component=='Disk':
                if self.get_spectral_component_by_name('Disk', verbose=False) is not None:
                    self._del_spectral_component('Disk', verbose=False)
                    self.EC_components_list.remove('Disk')

                if self.get_spectral_component_by_name('EC_Disk',verbose=False) is not None:
                    self._del_spectral_component('EC_Disk')
                    self._blob.do_EC_Disk = 0
                    self.EC_components_list.remove('EC_Disk')

                if self.get_spectral_component_by_name('EC_BLR', verbose=False) is not None:
                    self._blob.do_EC_BLR=0
                    self._del_spectral_component('EC_BLR', verbose=False)
                    self.EC_components_list.remove('EC_BLR')


            if EC_component=='EC_Disk':
                if self.get_spectral_component_by_name('EC_Disk', verbose=False) is not None:
                    self._blob.do_EC_Disk=0
                    self._del_spectral_component('EC_Disk', verbose=False)
                    self.EC_components_list.remove('EC_Disk')


            if EC_component=='EC_BLR':
                if self.get_spectral_component_by_name('EC_BLR', verbose=False) is not None:
                    self._blob.do_EC_BLR=0
                    self._del_spectral_component('EC_BLR', verbose=False)
                    self.EC_components_list.remove('EC_BLR')

            if EC_component=='DT':
                if self.get_spectral_component_by_name('DT', verbose=False) is not None:
                    self._del_spectral_component('DT', verbose=False)
                    self.EC_components_list.remove('DT')
                if self.get_spectral_component_by_name('EC_DT', verbose=False) is not None:
                    self._blob.do_EC_DT = 0
                    self._del_spectral_component('EC_DT', verbose=False)
                    self.EC_components_list.remove('EC_DT')

            if EC_component=='EC_DT':
                if self.get_spectral_component_by_name('EC_DT', verbose=False) is not None:
                    self._blob.do_EC_DT=0
                    self._del_spectral_component('EC_DT', verbose=False)
                    self.EC_components_list.remove('EC_DT')

            if EC_component=='EC_CMB':
                if self.get_spectral_component_by_name('EC_CMB', verbose=False) is not None:
                    self._blob.do_EC_CMB=0
                    self._del_spectral_component('EC_CMB', verbose=False)
                    self.EC_components_list.remove('EC_CMB')

            if EC_component=='Star':
                if self.get_spectral_component_by_name('Star', verbose=False) is not None:
                    self._blob.do_star=0
                    self._del_spectral_component('Star', verbose=False)
                    self.EC_components_list.remove('Star')

        self.del_par_from_dic(build_ExtFields_dic(EC_components_list,disk_type))



    def add_EC_component(self,EC_components_list=[],disk_type='BB'):

        if disk_type is not None:
            if disk_type not in self._allwed_disk_type:
                raise RuntimeError('disk type',disk_type,'not in allwowed', self._allwed_disk_type)
        else:
            if self.get_par_by_name('disk_type') is not None:
                disk_type=self.get_par_by_name('disk_type').val


        if isinstance(EC_components_list, six.string_types):
            EC_components_list=[EC_components_list]

        if 'All' in EC_components_list:
            EC_components_list=self._allowed_EC_components_list[::]

        for EC_component in EC_components_list:
            if EC_component not in self._allowed_EC_components_list:
                raise RuntimeError("EC_component %s not allowed" % EC_component, "please choose among ",
                                   self._allowed_EC_components_list)

            if EC_component == 'Disk':
                if self.get_spectral_component_by_name('Disk',verbose=False) is None:
                    self._add_spectral_component('Disk',var_name='do_Disk', state_dict=dict((('on', 1), ('off', 0))))
                    self.EC_components_list.append('Disk')

            if EC_component=='EC_Disk':
                #self._blob.do_EC_Disk=1
                if self.get_spectral_component_by_name('EC_Disk',verbose=False) is None:
                    self._add_spectral_component('EC_Disk', var_name='do_EC_Disk', state_dict=dict((('on', 1), ('off', 0))))
                    self.EC_components_list.append('EC_Disk')

                if self.get_spectral_component_by_name('Disk',verbose=False) is None:
                    self._add_spectral_component('Disk',var_name='do_Disk', state_dict=dict((('on', 1), ('off', 0))))
                    self.EC_components_list.append('Disk')

            if EC_component=='EC_BLR':
                #self._blob.do_EC_BLR=1
                if self.get_spectral_component_by_name('EC_BLR',verbose=False) is None:
                    self._add_spectral_component('EC_BLR', var_name='do_EC_BLR', state_dict=dict((('on', 1), ('off', 0))))
                    self.EC_components_list.append('EC_BLR')

                if self.get_spectral_component_by_name('Disk',verbose=False) is None:
                    self._add_spectral_component('Disk',var_name='do_Disk', state_dict=dict((('on', 1), ('off', 0))))
                    self.EC_components_list.append('Disk')


            if EC_component == 'DT':
                if self.get_spectral_component_by_name('DT',verbose=False) is None:
                    self._add_spectral_component('DT',var_name='do_DT', state_dict=dict((('on', 1), ('off', 0))))
                    self.EC_components_list.append('DT')

                if self.get_spectral_component_by_name('Disk',verbose=False) is None:
                    self._add_spectral_component('Disk',var_name='do_Disk', state_dict=dict((('on', 1), ('off', 0))))
                    self.EC_components_list.append('Disk')



            if EC_component == 'Star':
                if self.get_spectral_component_by_name('Star',verbose=False) is None:
                    self._add_spectral_component('Star',var_name='do_Star', state_dict=dict((('on', 1), ('off', 0))))
                    self.EC_components_list.append('Star')




            if EC_component=='EC_DT':
                #self._blob.do_EC_DT=1
                if self.get_spectral_component_by_name('EC_DT',verbose=False) is None:
                    self._add_spectral_component('EC_DT', var_name='do_EC_DT', state_dict=dict((('on', 1), ('off', 0))))
                    self.EC_components_list.append('EC_DT')

                if self.get_spectral_component_by_name('DT',verbose=False) is None:
                    self._add_spectral_component('DT',var_name='do_DT', state_dict=dict((('on', 1), ('off', 0))))
                    self.EC_components_list.append('DT')

                if self.get_spectral_component_by_name('Disk',verbose=False) is None:
                    self._add_spectral_component('Disk',var_name='do_Disk', state_dict=dict((('on', 1), ('off', 0))))
                    self.EC_components_list.append('Disk')

            if EC_component=='EC_CMB':
                #self._blob.do_EC_CMB=1
                if self.get_spectral_component_by_name('EC_CMB',verbose=False) is None:
                    self._add_spectral_component('EC_CMB', var_name='do_EC_CMB', state_dict=dict((('on', 1), ('off', 0))))
                    self.EC_components_list.append('EC_CMB')

        #IF disk_type is already a parameter it has to be updated here to make it effective
        self.parameters.add_par_from_dict(build_ExtFields_dic(self.EC_components_list, disk_type),self,'_blob',JetParameter)
        #print('--> disk_type',disk_type)
        if disk_type is not None:

            #remove the old disk type parameters
            if  self.parameters.get_par_by_name('disk_type') is not None:
                self.del_par_from_dic(build_ExtFields_dic(['Disk'],self.parameters.get_par_by_name('disk_type').val))
            else:
                self.EC_components_list.append('Disk')
            self.parameters.add_par_from_dict(build_ExtFields_dic(self.EC_components_list, disk_type),self,'_blob',JetParameter)
            self.parameters.disk_type.val = disk_type



    def del_par_from_dic(self,model_dic):
        """
        """
        for key in model_dic.keys():

            par=self.parameters.get_par_by_name(key)

            if par is not None:
                self.parameters.del_par(par)



    def get_DL_cm(self,eval=False):

        if eval is True:
            self.set_blob()

        if self.cosmo._c is not None:
            return self.cosmo.get_DL_cm(self.parameters.z_cosm.val)
        else:
            return self.cosmo.get_DL_cm()


    @safe_run
    def get_beaming(self,):

        BlazarSED.SetBeaming(self._blob)
        return self._blob.beam_obj


    def set_flag(self,flag):
        self._blob.STEM=flag

    def get_flag(self):
        return self._blob.STEM

    def get_path(self):
        return self._blob.path

    def set_path(self,path,clean_work_dir=True):
        if path.endswith('/'):
            pass
        else:
            path+='/'

        set_str_attr(self._blob,'path',path)
        makedir(path,clean_work_dir=clean_work_dir)

    def get_IC_mode(self):
        return dict(map(reversed, self._IC_states.items()))[self._blob.do_IC]



    def set_external_field_transf(self,val):
        if val not in self._external_field_transf.keys():
            raise RuntimeError('val',val,'not in allowed values',self._external_field_transf.keys())

        self._blob.EC_stat=self._external_field_transf[val]

    def get_external_field_transf(self):
        return dict(map(reversed, self._external_field_transf.items()))[self._blob.EC_stat]

    def set_emiss_lim(self,val):
        self._blob.emiss_lim=val

    def get_emiss_lim(self):
        return self._blob.emiss_lim


    @property
    def IC_nu_size(self):
        return self._blob.nu_IC_size

    @IC_nu_size.setter
    def IC_nu_size(self, val):
        self.set_IC_nu_size(val)

    def get_IC_nu_size(self):
        return self._blob.nu_IC_size

    def set_IC_nu_size(self, val):
        if val > self._nu_static_size:
            raise RuntimeError('value can not exceed',self._nu_static_size)
        self._blob.nu_IC_size = val

    @property
    def nu_seed_size(self):
        return self._blob.nu_seed_size

    def get_seed_nu_size(self):
        return self._blob.nu_seed_size

    @nu_seed_size.setter
    def nu_seed_size(self,val):
        self.set_seed_nu_size(val)

    def set_seed_nu_size(self,val):
        if val>self._nu_static_size:
            raise RuntimeError('value can not exceed',self._nu_static_size)
        self._blob.nu_seed_size=val



    def set_gamma_grid_size(self,val):
        self.emitters_distribution.set_grid_size(gamma_grid_size=val)

    @property
    def gamma_grid_size(self):
        return self._blob.gamma_grid_size

    @gamma_grid_size.setter
    def gamma_grid_size(self,val):
        self.emitters_distribution.set_grid_size(gamma_grid_size=val)

    @property
    def nu_min(self):
        return self._get_nu_min_grid()

    @nu_min.setter
    def nu_min(self, val):
        if hasattr(self, '_blob'):
            self._set_nu_min_grid(val)

    def _get_nu_min_grid(self):
        return  self._blob.nu_start_grid


    def _set_nu_min_grid(self, val):
        self._blob.nu_start_grid=val

    @property
    def Norm_distr(self):
        if hasattr(self,'emitters_distribution'):
            if self.emitters_distribution._user_defined is False:
                return self._blob.Norm_distr
            else:
                return self.emitters_distribution.normalize
        else:
            return None


    @Norm_distr.setter
    def Norm_distr(self, val):
        if hasattr(self, 'emitters_distribution'):

                if val == 1 or val is True:
                    if self.emitters_distribution._user_defined is False:
                        self._blob.Norm_distr = 1
                    else:
                        self.emitters_distribution.normalize = val
                    self.parameters.N.par_type='emitters_density'
                    self.parameters.N.units ='1/cm3'
                elif val == 0 or val is False:
                    if self.emitters_distribution._user_defined is False:
                        self._blob.Norm_distr = 0
                    else:
                        self.emitters_distribution.normalize = val
                        self.parameters.N.par_type = 'scaling_factor'
                else:
                    raise RuntimeError('value', val, 'not allowed, allowed 0/1 or False/True')




    def switch_Norm_distr_ON(self):
        self.Norm_distr=True

    def switch_Norm_distr_OFF(self):
        self.Norm_distr = False

    @property
    def nu_max(self):
        return self._get_nu_max_grid()

    @nu_max.setter
    def nu_max(self, val):
        if hasattr(self, '_blob'):
            self._set_nu_max_grid(val)

    def _set_nu_max_grid(self, val):
        self._blob.nu_stop_grid=val

    def _get_nu_max_grid(self):
        return  self._blob.nu_stop_grid

    @property
    def nu_size(self):
        return self._nu_size

    @nu_size.setter
    def nu_size(self, size):
        self._nu_size = size

    def set_nu_grid_size(self, val):
        self._set_nu_grid_size_blob(val)

    @property
    def nu_grid_size(self):
        return self._get_nu_grid_size_blob()

    @nu_grid_size.setter
    def nu_grid_size(self, val):
        self._set_nu_grid_size_blob(val)


    def _set_nu_grid_size_blob(self, val):
        if val < 100:
            val = 100
        if val > self._static_spec_arr_grid_size:
            raise RuntimeError('value can not exceed', self._nu_static_size)
        self._blob.nu_grid_size=val

    def _get_nu_grid_size_blob(self):
        return  self._blob.nu_grid_size



    def set_verbosity(self,val):
        self._blob.verbose=val

    def get_verbosity(self):
        return  self._blob.verbose

    def debug_synch(self):
        print ("nu stop synch", self._blob.nu_stop_Sync)
        print ("nu stop synch ssc", self._blob.nu_stop_Sync_ssc)
        print ("ID MAX SYNCH", self._blob.NU_INT_STOP_Sync_SSC)

    def debug_SSC(self):
        print ("nu start SSC", self._blob.nu_start_SSC)
        print ("nu stop SSC", self._blob.nu_stop_SSC)
        print ("ID MAX SSC", self._blob.NU_INT_STOP_COMPTON_SSC)


    def show_emitters_distribution(self):
        print('-'*80)
        print('%s distribution:'%self._emitters_type)
        print(" type: %s  " % (self._emitters_distribution_name))
        print(" gamma energy grid size: ", self.gamma_grid_size)
        print(" gmin grid : %e" % self._blob.
              gmin_griglia)
        print(" gmax grid : %e" % self._blob.gmax_griglia)
        print(" normalization ", self.Norm_distr)
        print(" log-values ", self._emitters_distribution_log_values)
        print('')
        self.parameters.par_array.sort(key=lambda x: x.name, reverse=False)

        self.parameters.show_pars(names_list=self._emitters_distribution_dic.keys())

    def show_model(self):
        """
        shortcut to :class:`ModelParametersArray.show_pars` method
        shows all the paramters in the model

        """
        print("")
        print('-'*80)

        print ("jet model description")
        print('-'*80)
        print("name: %s  " % (self.name))
        print('')
        print('%s distribution:'%self._emitters_type)
        print(" type: %s  " % (self._emitters_distribution_name))
        print (" gamma energy grid size: ",self.gamma_grid_size)
        print (" gmin grid : %e"%self._blob.gmin_griglia)
        print (" gmax grid : %e"%self._blob.gmax_griglia)
        print(" normalization ", self.Norm_distr)
        print(" log-values ", self._emitters_distribution_log_values)
        print('')
        if 'Disk' in self.EC_components_list:
            print('accretion disk:')
            BlazarSED.set_Disk(self._blob)
            print(' disk Type: %s'%self.parameters.get_par_by_name('disk_type').val)
            print(' L disk: %e (erg/s)'%self.parameters.get_par_by_name('L_Disk').val)
            print(' T disk: %e (K)'%self._blob.T_Disk)
            print(' nu peak disk: %e (Hz)'%BlazarSED.eval_nu_peak_Disk(self._blob.T_Disk))
            if self.parameters.get_par_by_name('disk_type').val == 'MultiBB':
                print(' Sw radius %e (cm)'%self._blob.R_Sw)
                print(' L Edd. %e (erg/s)'%self._blob.L_Edd)
                yr=86400*365
                print(' accr_rate: %e (M_sun/yr)'%(yr*self._blob.accr_rate/BlazarSED.m_sun))
                print(' accr_rate Edd.: %e (M_sun/yr)'%(yr*self._blob.accr_Edd/BlazarSED.m_sun))


            print('')
        print('radiative fields:')
        print (" seed photons grid size: ", self.nu_seed_size)
        print (" IC emission grid size: ", self.get_IC_nu_size())
        print (' source emissivity lower bound :  %e' % self._blob.emiss_lim)
        print (' spectral components:')
        for _s in self.spectral_components_list:
            print("   name:%s,"%_s.name, 'state:', _s.state)
        print('external fields transformation method:', self.get_external_field_transf())
        print ('')
        print ('SED info:')
        print (' nu grid size jetkernel: %d' % self.nu_grid_size)
        print (' nu size: %d' % self.nu_size)
        print (' nu mix (Hz): %e' % self._get_nu_min_grid())
        print (' nu max (Hz): %e' % self._get_nu_max_grid())
        print('')
        print('flux plot lower bound   :  %e' % self.flux_plot_lim)
        print('')
        print('-'*80)
        self.show_pars()
        print('-' * 80)

    def plot_model(self,plot_obj=None,clean=False,label=None,comp=None,sed_data=None,color=None,auto_label=True,line_style='-',frame='obs', density=False):
        plot_obj=self._set_up_plot(plot_obj,sed_data,frame,density)

        if clean is True:
            plot_obj.clean_model_lines()

        if comp is not None:
            c = self.get_spectral_component_by_name(comp)

            if label is not None:
                comp_label = label
            elif label is None and auto_label is True:
                comp_label = c.name
            else:
                comp_label=None
            if c.state!='off':
                plot_obj.add_model_plot(c.SED, line_style=line_style, label=comp_label,flim=self.flux_plot_lim,color=color,auto_label=auto_label, density=density,update=False, frame=frame)

        else:
            for c in self.spectral_components_list:
                comp_label = c.name
                if auto_label is not True:
                    comp_label=label
                #print('comp label',comp_label)
                if c.state != 'off' and c.name!='Sum':
                    plot_obj.add_model_plot(c.SED, line_style=line_style, label=comp_label,flim=self.flux_plot_lim,auto_label=auto_label,color=color, density=density,update=False, frame=frame)

            c=self.get_spectral_component_by_name('Sum')
            if label is not None:
                comp_label = label
            else:
                comp_label='Sum'

            plot_obj.add_model_plot(c.SED, line_style='--', label=comp_label, flim=self.flux_plot_lim,color=color, density=density,update=False)

        plot_obj.update_plot()

        return plot_obj

    @safe_run
    def set_blob(self):
        if hasattr(self, 'T_esc_e_second'):
            if self.T_esc_e_second is None:
                self._blob.T_esc_e_second = self.parameters.R.val / BlazarSED.vluce_cm
            else:
                self._blob.T_esc_e_second = self.T_esc_e_second
        if self.emitters_distribution._user_defined is True:
            self.emitters_distribution._fill()
        BlazarSED.Init(self._blob, self.get_DL_cm())
        if self.emitters_distribution._user_defined is True:
            self.emitters_distribution._set_blob()


    @safe_run
    def set_external_fields(self):
        self.set_blob()
        BlazarSED.spectra_External_Fields(1,self._blob)

    def lin_func(self, lin_nu, init, phys_output=False, update_emitters=True):
        if self.emitters_distribution is None:
            raise RuntimeError('emitters distribution not defined')

        if init is True:
            self.set_blob()
            self._update_spectral_components()
        BlazarSED.Run_SED(self._blob)

        if phys_output==True:
            BlazarSED.EnergeticOutput(self._blob,0)

        if self.emitters_distribution._user_defined is False:
            if update_emitters is True:
                self.emitters_distribution.update()
            else:
                self.emitters_distribution._fill()

        nu_sed_sum, nuFnu_sed_sum = self.spectral_components.Sum.get_SED_points(lin_nu=lin_nu,log_log=False,interp=self._jetkernel_interp)
        return nu_sed_sum, nuFnu_sed_sum

    def _eval_model(self, lin_nu, log_nu, init, loglog, phys_output=False, update_emitters=True):
        log_model = None
        lin_nu, lin_model = self.lin_func(lin_nu, init, phys_output, update_emitters)
        if loglog is True:
            log_model = np.log10(lin_model)

        return lin_model, log_model


    def _prepare_nu_model(self, nu, loglog):
        if nu is None:
            lin_nu = np.logspace(np.log10(self.nu_min), np.log10(self.nu_max), self.nu_size)
            log_nu = np.log10(lin_nu)
        else:
            if np.shape(nu) == ():
                nu = np.array([nu])

            if loglog is True:
                lin_nu = np.power(10., nu)
                log_nu = nu
            else:
                log_nu = np.log10(nu)
                lin_nu = nu

        return lin_nu, log_nu

    @safe_run
    def eval(self,
             init=True,
             fill_SED=True,
             nu=None,
             get_model=False,
             loglog=False,
             plot=None,
             label=None,
             phys_output=False,
             update_emitters=True):
        out_model = None
        lin_nu, log_nu = self._prepare_nu_model(nu, loglog)

        lin_model, log_model= self._eval_model(lin_nu, log_nu ,init, loglog, phys_output=phys_output,
                                                update_emitters=update_emitters)
        #print('-->',lin_nu.min(),lin_nu.max())
        if fill_SED is True:
            self._fill(lin_nu,lin_model)

        if get_model is True:

            if loglog is True:
                out_model = log_model
            else:
                out_model = lin_model

        return out_model


    def energetic_report(self,verbose=True):
        self._build_energetic_report()            
        if verbose is True:
            _show_table(self.energetic_report_table)

    def _build_energetic_report(self,):
        self.energetic_dict={}
        BlazarSED.SetBeaming(self._blob)
        _energetic = BlazarSED.EnergeticOutput(self._blob,0)
        _par_array=ModelParameterArray()

        _name = [i for i in _energetic.__class__.__dict__.keys() if i[:1] != '_']
        _par_array.add_par(ModelParameter(name='BulkLorentzFactor', val=self._blob.BulkFactor, units='',par_type=''))
        self.energetic_dict['BulkLorentzFactor']= self._blob.BulkFactor
        try:
            for _n in _name:
                units = 'skip_this'
                if _n[0]=='L':
                    par_type='Lum. blob rest. frme.'
                    units='erg/s'
                elif _n[0] == 'U' and 'DRF' not in _n:
                    par_type = 'Energy dens. blob rest. frame'
                    units = 'erg/cm^3'
                elif _n[0] == 'U' and 'DRF'  in _n:
                    par_type = 'Energy dens. disk rest. frame'
                    units = 'erg/cm^3'
                elif _n[0] == 'j':
                    par_type = 'jet Lum.'
                    units = 'erg/s'
                else:
                    if _n !='thisown':
                        warnings.warn('energetic name %s not understood'%_n)

                if units == 'skip_this':
                    pass
                else:
                    self.energetic_dict[_n]=getattr(_energetic, _n)

                    _par_array.add_par(ModelParameter(name=_n, val=getattr(_energetic, _n), units=units,par_type=par_type))


            if self.emitters_distribution.emitters_type=='electrons':
                _i=_par_array.par_table['name']=='U_p_target'
                _par_array.par_table.remove_rows(_i)
                _=self.energetic_dict.pop('U_p_target')

            if self.emitters_distribution.emitters_type=='protons':
                _i=_par_array.par_table['name']=='U_p_cold'
                _par_array.par_table.remove_rows(_i)
                _=self.energetic_dict.pop('U_p_cold')

            self.energetic_report_table = _par_array.par_table
            self.energetic_report_table.remove_columns(['log','frozen','phys. bound. min','phys. bound. max'])
            self.energetic_report_table.rename_column('par type','type')

        except Exception as e:
            print('_energetic',_energetic)
            raise RuntimeError('energetic_report failed',e)







    def get_SED_peak(self,peak_name=None,freq_range=None,log_log=False):

        if peak_name is not None and freq_range is not None:
            print ("either you provide peak_name or freq_range")
            raise ValueError
        elif   peak_name is not None:
            try:
                if log_log==False:
                    return getattr(self._blob,peak_name)
                else:
                     return np.log10(getattr(self._blob,peak_name) )
            except:
                print ("peak name %s, not found, check name"%peak_name)
                raise ValueError

        else:
            x,y=self.spectral_components.Sum.get_SED_points(log_log=log_log)
            msk1=x>freq_range[0]
            msk2=x<freq_range[1]
            y_m= y[msk1*msk2].max()
            x_id= np.argmax(y[msk1*msk2])
            return x[msk1*msk2][x_id],y_m

    def get_component_peak(self,comp_name=None,log_log=False):
        comp = self.get_spectral_component_by_name(comp_name)

        ID = np.argmax(comp.SED.nuFnu.value)
        x_p, y_p = comp.SED.nu[ID].value, comp.SED.nuFnu[ID].value

        if log_log is True:
            x_p, y_p=np.log10([x_p,y_p])

        return x_p, y_p




class Jet(JetBase):
    """ Jet class

    """

    def __init__(self,
                 cosmo=None,
                 name=None,
                 emitters_type='electrons',
                 emitters_distribution='plc',
                 emitters_distribution_log_values=False,
                 beaming_expr='delta',
                 T_esc_e_second=None,
                 jet_workplace=None,
                 verbose=None,
                 clean_work_dir=True,
                 electron_distribution=None,
                 proton_distribution=None,
                 electron_distribution_log_values=None,
                 proton_distribution_log_values=None):
        """

        Parameters
        ----------
        cosmo
        name
        emitters_type
        emitters_distribution
        emitters_distribution_log_values
        beaming_expr
        T_esc_e_second
        jet_workplace
        verbose
        clean_work_dir
        electron_distribution
        proton_distribution
        electron_distribution_log_values
        proton_distribution_log_values
        """
        if electron_distribution is not None:
            emitters_type = 'electrons'
            emitters_distribution= electron_distribution
        if electron_distribution_log_values is not None:
            emitters_distribution_log_values = electron_distribution_log_values

        if proton_distribution is not None:
            emitters_type = 'protons'
            emitters_distribution= proton_distribution
        if proton_distribution_log_values is not None:
            emitters_distribution_log_values = proton_distribution_log_values


        if name is None:
            _name = 'jet'
        else:
            _name = name

        super(Jet,self).__init__(cosmo=cosmo,
                                 name=_name,
                                 emitters_type=emitters_type,
                                 emitters_distribution=emitters_distribution,
                                 emitters_distribution_log_values=emitters_distribution_log_values,
                                 beaming_expr=beaming_expr,
                                 jet_workplace=jet_workplace,
                                 verbose=verbose,
                                 clean_work_dir=clean_work_dir)

        if name is None or name == '':
            if self.emitters_distribution.emitters_type == 'electrons':
                name = 'jet_leptonic'
            elif self.emitters_distribution.emitters_type == 'protons':
                name = 'jet_hadronic_pp'
            else:
                name = 'jet'

        self.name = clean_var_name(name)

        if self.emitters_distribution.emitters_type == 'protons':
            self.T_esc_e_second=T_esc_e_second
            self.add_pp_gamma_component()
            self.add_pp_neutrino_component()
            self.add_bremss_ep_component()

        if self.emitters_distribution.emitters_type == 'electrons':
            self.electron_distribution=self.emitters_distribution

        if self.emitters_distribution.emitters_type == 'protons':
            self.protons_distribution=self.emitters_distribution

    @staticmethod
    def available_electron_distributions():
        JetBase.available_emitters_distributions()

    @staticmethod
    def available_proton_distributions():
        JetBase.available_emitters_distributions()

    def get_proton_distribution_name(self):
        return self.get_emitters_distribution_name()

    def get_electron_distribution_name(self):
        return self.get_emitters_distribution_name()

    def show_proton_distribution(self):
        self.show_emitters_distribution()

    def show_electron_distribution(self):
        self.show_emitters_distribution()

    def add_bremss_ep_component(self):
        self._add_spectral_component('Bremss_ep', var_name='do_bremss_ep', state_dict=dict((('on', 1), ('off', 0))))

    def add_pp_gamma_component(self):
        self._add_spectral_component('PP_gamma', var_name='do_pp_gamma', state_dict=dict((('on', 1), ('off', 0))))

    def add_pp_neutrino_component(self):
        self._add_spectral_component('PP_neutrino_tot', var_name='do_pp_neutrino',
                                     state_dict=dict((('on', 1), ('off', 0))))
        self._add_spectral_component('PP_neutrino_mu', var_name='do_pp_neutrino',
                                     state_dict=dict((('on', 1), ('off', 0))))
        self._add_spectral_component('PP_neutrino_e', var_name='do_pp_neutrino',
                                     state_dict=dict((('on', 1), ('off', 0))))





    def set_N_from_U_emitters(self,U, gmin=None, gmax=None):
        """ Sets the normalization of N to match the energy density of the primary emitters
        Parameters
        ----------
        U: float, (erg/cm3)
        gmin: float, optional,
            minimum value to evaluate the integral

        gmax: float, optional,
            maximum value to evaluate the integral
        Returns
        -------

        """
        N = self.parameters.N.val
        #gamma_grid_size = self._blob.gamma_grid_size
        #self.emitters_distribution.set_grid_size(1000)
        #self.emitters_distribution._fill()
        #self.set_blob()
        #BlazarSED.EvalU_e(self._blob)
        ratio = U/self.emitters_distribution.eval_U(gmin=gmin, gmax=gmax)
        #self.emitters_distribution.set_grid_size(gamma_grid_size)
        self.emitters_distribution._fill()
        self.set_par('N', val=N *ratio)
        self.set_blob()

    def set_N_from_U_vol_emitters(self, U_vol, gmin=None, gmax=None):
        """Sets the normalization of N to match the volume integrated energy of the primary emitters

        Parameters
        ----------
        U_vol: float (erg)

        gmin: float, optional,
            minimum value to evaluate the integral

        gmax: float, optional,
            maximum value to evaluate the integral

        Returns
        -------

        """
        #gamma_grid_size = self._blob.gamma_grid_size
        #self.emitters_distribution.set_grid_size(1000)
        #self.set_blob()
        U=U_vol/ self._blob.Vol_sphere
        #self.emitters_distribution.set_grid_size(gamma_grid_size)
        self.set_N_from_U_emitters(U, gmin=gmin, gmax=gmax)


    def set_N_from_L_sync(self,L_sync):
        """Sets the normalization of N to match the src integrated Luminosity of the   synchrotron emission

        Parameters
        ----------
        L_sync : float (erg/s)

        Returns
        -------

        """
        self.set_par('N', val=1.0)
        #gamma_grid_size = self._blob.gamma_grid_size
        #self.emitters_distribution.set_grid_size(100)
        #self.set_blob()
        delta = self._blob.beam_obj
        ratio = L_sync/(BlazarSED.Power_Sync_Electron(self._blob)* delta ** 4)
        #self.emitters_distribution.set_grid_size(gamma_grid_size)
        self.set_par('N', val=ratio)


    def set_N_from_F_sync(self, F_sync):
        """Sets the normalization of N to match the observed integrated synchrotron flux

        Parameters
        ----------
        F_sync : float, (erg cm-1 s-1 Hz-1)
            observed integrated synchrotron flux

        Returns
        -------

        """
        DL = self.get_DL_cm()
        L = F_sync * DL * DL * 4.0 * np.pi
        self.set_N_from_L_sync(L)

    def set_N_from_nuLnu(self,nuLnu_src, nu_src):
        """Sets the normalization of N to match the src Luminosity of the   synchrotron emission at src frequency nu

        Parameters
        ----------
        nuLnu_src : float, (erg/s)
            Luminosity of the   synchrotron emission at src frequency nu

        nu_src: float (Hz)
            synchrotron emission at src frequency nu (Hz)

        Returns
        -------

        """
        self.set_par('N',val=1.0)
        #gamma_grid_size = self._blob.gamma_grid_size
        #self.emitters_distribution.set_grid_size(100)
        self.set_blob()
        delta = self._blob.beam_obj
        nu_blob = nu_src / delta
        L_out = BlazarSED.Lum_Sync_at_nu(self._blob, nu_blob) * delta ** 4
        N_out = nuLnu_src / L_out
        #self.emitters_distribution.set_grid_size(gamma_grid_size)
        self.set_par('N', val=N_out)


    def set_N_from_nuFnu(self, nuFnu_obs, nu_obs):
        """Sets the normalization of N to match the observed flux nuFnu_obs at a given frequency nu_obs

        Parameters
        ----------
        nuFnu_obs: float, (erg cm-1 s-1 Hz-1)
            observed differential synchrotron flux

        nu_obs: float, (Hz)
            synchrotron emission at src frequency nu (Hz)

        Returns
        -------

        """

        self.set_blob()
        DL =  self.get_DL_cm()
        L = nuFnu_obs * DL * DL * 4.0 * np.pi
        nu_rest = nu_obs * (1 + self.parameters.z_cosm.val)
        self.set_N_from_nuLnu( L, nu_rest)



    def set_B_eq(self, nuFnu_obs, nu_obs, B_min=1E-9,B_max=1.0,N_pts=20,plot=False):
        """Sets the magnetic field (B) equipartition from numerical minimization over a logarithmic grid  of B values,
        for a given observed flux of the  synchrotron emission (nuFnu_obs) at a given observed frequency (nu_obs)

        Parameters
        ----------
        nuFnu_obs: float, (erg cm-1 s-1 Hz-1)
            observed differential synchrotron flux

        nu_obs: float, (Hz)
            synchrotron emission at src frequency nu (Hz)

        B_min: float, (Gauss), optional
            lower bound for B-grid

        B_max:  float, (Gauss), optional
            upper bound for B-grid

        N_pts: int, optional
            Other number of points to build the B-grid

        plot: book, optional, default=False
            if True plots the numerical grid

        Returns
        -------

        """

        b_grid = np.logspace(np.log10(B_min), np.log10(B_max), N_pts)
        print ('B grid min ',B_min)
        print ('B grid max ',B_max)
        print ('grid points',N_pts)
        U_e = np.zeros(N_pts)
        U_B = np.zeros(N_pts)
        N = np.zeros(N_pts)
        self.set_par('B', b_grid[0])
        self.set_blob()

        for ID, b in enumerate(b_grid):
            self.set_par('B', b)
            self.set_par('N', 1.0)
            # print 'B_eq',ID
            self.set_N_from_nuFnu(nuFnu_obs, nu_obs)
            N[ID]=self.get_par_by_name('N').val
            self.set_blob()
            #
            U_e[ID] = self._blob.U_e
            U_B[ID] = self._blob.UB
            # delta=Jet.get_beaming()
            # print "check L_in=%4.4e L_out=%4.4e"%(L_0,(L_0/delta**4)/BlazarSED.Power_Sync_Electron(Jet._Jet__blob))

        ID_min = np.argmin(U_B + U_e)

        if plot==True:
            #import  pylab as plt
            fig, ax = plt.subplots()
            ax.plot(b_grid,U_e)
            ax.plot(b_grid,U_B)
            ax.plot(b_grid,U_B+U_e)
            ax.scatter(b_grid[ID_min],U_e[ID_min]+U_B[ID_min])

            ax.semilogy()
            ax.semilogx()
            plt.show()


        self.set_par('B', val=b_grid[ID_min])
        self.set_par('N', val=N[ID_min])
        print ('setting B to ',b_grid[ID_min])
        print ('setting N to ',N[ID_min])
        return b_grid[ID_min],b_grid,U_B,U_e

